# Copyright 2023 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from typing import Optional
import unittest
from unittest import mock

import docker
from kfp import dsl
from kfp import local
from kfp.dsl import Artifact
from kfp.dsl import Output
from kfp.local import docker_task_handler
from kfp.local import testing_utilities


class DockerMockTestCase(unittest.TestCase):

    def setUp(self):
        super().setUp()
        self.docker_mock = mock.Mock()
        patcher = mock.patch('docker.from_env')
        self.mocked_docker_client = patcher.start().return_value

        mock_container = mock.Mock()
        self.mocked_docker_client.containers.run.return_value = mock_container
        # mock successful run
        mock_container.logs.return_value = [
            'fake'.encode('utf-8'),
            'container'.encode('utf-8'),
            'logs'.encode('utf-8'),
        ]
        mock_container.wait.return_value = {'StatusCode': 0}

    def teardown(self):
        super().tearDown()
        self.docker_mock.reset_mock()


class TestRunDockerContainer(DockerMockTestCase):

    def test_no_volumes(self):
        docker_task_handler.run_docker_container(
            docker.from_env(),
            image='alpine',
            command=['echo', 'foo'],
            volumes={},
        )

        self.mocked_docker_client.containers.run.assert_called_once_with(
            image='alpine:latest',
            command=['echo', 'foo'],
            detach=True,
            stdout=True,
            stderr=True,
            volumes={},
        )

    def test_cwd_volume(self):
        current_test_dir = os.path.dirname(os.path.abspath(__file__))
        docker_task_handler.run_docker_container(
            client=docker.from_env(),
            image='alpine',
            command=['cat', '/localdir/docker_task_handler_test.py'],
            volumes={current_test_dir: {
                'bind': '/localdir',
                'mode': 'ro'
            }},
        )
        self.mocked_docker_client.containers.run.assert_called_once_with(
            image='alpine:latest',
            command=['cat', '/localdir/docker_task_handler_test.py'],
            detach=True,
            stdout=True,
            stderr=True,
            volumes={current_test_dir: {
                'bind': '/localdir',
                'mode': 'ro'
            }})


class TestDockerTaskHandler(DockerMockTestCase):

    def test_get_volumes_to_mount(self):
        handler = docker_task_handler.DockerTaskHandler(
            image='alpine',
            full_command=['echo', 'foo'],
            pipeline_root=os.path.abspath('my_root'),
            runner=local.DockerRunner(),
        )
        volumes = handler.get_volumes_to_mount()
        self.assertEqual(
            volumes, {
                os.path.abspath('my_root'): {
                    'bind': os.path.abspath('my_root'),
                    'mode': 'rw'
                }
            })

    def test_run(self):
        handler = docker_task_handler.DockerTaskHandler(
            image='alpine',
            full_command=['echo', 'foo'],
            pipeline_root=os.path.abspath('my_root'),
            runner=local.DockerRunner(),
        )

        handler.run()
        self.mocked_docker_client.containers.run.assert_called_once_with(
            image='alpine:latest',
            command=['echo', 'foo'],
            detach=True,
            stdout=True,
            stderr=True,
            volumes={
                os.path.abspath('my_root'): {
                    'bind': os.path.abspath('my_root'),
                    'mode': 'rw'
                }
            },
        )

    def test_pipeline_root_relpath(self):
        with self.assertRaisesRegex(
                ValueError,
                r"'pipeline_root' should be an absolute path to correctly construct the volume mount specification\."
        ):
            docker_task_handler.DockerTaskHandler(
                image='alpine',
                full_command=['echo', 'foo'],
                pipeline_root='my_relpath',
                runner=local.DockerRunner(),
            ).run()


class TestAddLatestTagIfNotPresent(unittest.TestCase):

    def test_no_tag(self):
        actual = docker_task_handler.add_latest_tag_if_not_present(
            image='alpine')
        expected = 'alpine:latest'
        self.assertEqual(actual, expected)

    def test_latest_tag(self):
        actual = docker_task_handler.add_latest_tag_if_not_present(
            image='alpine:latest')
        expected = 'alpine:latest'
        self.assertEqual(actual, expected)

    def test_no_tag(self):
        actual = docker_task_handler.add_latest_tag_if_not_present(
            image='alpine:3.19.0')
        expected = 'alpine:3.19.0'
        self.assertEqual(actual, expected)


class TestE2E(DockerMockTestCase,
              testing_utilities.LocalRunnerEnvironmentTestCase):

    def setUp(self):
        super().setUp()
        local.init(runner=local.DockerRunner())

    def test_python(self):

        @dsl.component
        def artifact_maker(x: str, a: Output[Artifact]):
            with open(a.path, 'w') as f:
                f.write(x)

        try:
            artifact_maker(x='foo')
        # cannot get outputs if they aren't created due to mock
        except FileNotFoundError:
            pass

        run_mock = self.mocked_docker_client.containers.run
        run_mock.assert_called_once()
        kwargs = run_mock.call_args[1]
        self.assertEqual(
            kwargs['image'],
            'python:3.9',
        )
        self.assertTrue(
            any('def artifact_maker' in c for c in kwargs['command']))
        self.assertTrue(kwargs['detach'])
        self.assertTrue(kwargs['stdout'])
        self.assertTrue(kwargs['stderr'])
        root_vol_key = [
            key for key in kwargs['volumes'].keys() if 'local_outputs' in key
        ][0]
        self.assertEqual(kwargs['volumes'][root_vol_key]['bind'], root_vol_key)
        self.assertEqual(kwargs['volumes'][root_vol_key]['mode'], 'rw')

    def test_empty_container_component(self):

        @dsl.container_component
        def comp():
            return dsl.ContainerSpec(image='alpine')

        try:
            comp()
        # cannot get outputs if they aren't created due to mock
        except FileNotFoundError:
            pass

        run_mock = self.mocked_docker_client.containers.run
        run_mock.assert_called_once()
        kwargs = run_mock.call_args[1]
        self.assertEqual(
            kwargs['image'],
            'alpine:latest',
        )
        self.assertEqual(kwargs['command'], [])

    def test_container_component(self):

        @dsl.container_component
        def artifact_maker(x: str,):
            return dsl.ContainerSpec(
                image='alpine',
                command=['sh', '-c', f'echo prefix-{x}'],
            )

        try:
            artifact_maker(x='foo')
        # cannot get outputs if they aren't created due to mock
        except FileNotFoundError:
            pass

        run_mock = self.mocked_docker_client.containers.run
        run_mock.assert_called_once()
        kwargs = run_mock.call_args[1]
        self.assertEqual(
            kwargs['image'],
            'alpine:latest',
        )
        self.assertEqual(kwargs['command'], [
            'sh',
            '-c',
            'echo prefix-foo',
        ])
        self.assertTrue(kwargs['detach'])
        self.assertTrue(kwargs['stdout'])
        self.assertTrue(kwargs['stderr'])
        root_vol_key = [
            key for key in kwargs['volumes'].keys() if 'local_outputs' in key
        ][0]
        self.assertEqual(kwargs['volumes'][root_vol_key]['bind'], root_vol_key)
        self.assertEqual(kwargs['volumes'][root_vol_key]['mode'], 'rw')

    def test_if_present_with_string_omitted(self):

        @dsl.container_component
        def comp(x: Optional[str] = None):
            return dsl.ContainerSpec(
                image='alpine:3.19.0',
                command=[
                    dsl.IfPresentPlaceholder(
                        input_name='x',
                        then=['echo', x],
                        else_=['echo', 'No input provided!'])
                ])

        comp()

        run_mock = self.mocked_docker_client.containers.run
        run_mock.assert_called_once()
        kwargs = run_mock.call_args[1]
        self.assertEqual(
            kwargs['image'],
            'alpine:3.19.0',
        )
        self.assertEqual(kwargs['command'], [
            'echo',
            'No input provided!',
        ])

    def test_if_present_with_string_provided(self):

        @dsl.container_component
        def comp(x: Optional[str] = None):
            return dsl.ContainerSpec(
                image='alpine:3.19.0',
                command=[
                    dsl.IfPresentPlaceholder(
                        input_name='x',
                        then=['echo', x],
                        else_=['echo', 'No artifact provided!'])
                ])

        comp(x='foo')

        run_mock = self.mocked_docker_client.containers.run
        run_mock.assert_called_once()
        kwargs = run_mock.call_args[1]
        self.assertEqual(
            kwargs['image'],
            'alpine:3.19.0',
        )
        self.assertEqual(kwargs['command'], [
            'echo',
            'foo',
        ])

    def test_if_present_single_element_with_string_omitted(self):

        @dsl.container_component
        def comp(x: Optional[str] = None):
            return dsl.ContainerSpec(
                image='alpine:3.19.0',
                command=[
                    'echo',
                    dsl.IfPresentPlaceholder(
                        input_name='x',
                        then=x,
                        else_='No artifact provided!',
                    )
                ])

        comp()

        run_mock = self.mocked_docker_client.containers.run
        run_mock.assert_called_once()
        kwargs = run_mock.call_args[1]
        self.assertEqual(
            kwargs['image'],
            'alpine:3.19.0',
        )
        self.assertEqual(kwargs['command'], [
            'echo',
            'No artifact provided!',
        ])

    def test_if_present_single_element_with_string_provided(self):

        @dsl.container_component
        def comp(x: Optional[str] = None):
            return dsl.ContainerSpec(
                image='alpine:3.19.0',
                command=[
                    'echo',
                    dsl.IfPresentPlaceholder(
                        input_name='x',
                        then=x,
                        else_='No artifact provided!',
                    )
                ])

        comp(x='foo')

        run_mock = self.mocked_docker_client.containers.run
        run_mock.assert_called_once()
        kwargs = run_mock.call_args[1]
        self.assertEqual(
            kwargs['image'],
            'alpine:3.19.0',
        )
        self.assertEqual(kwargs['command'], [
            'echo',
            'foo',
        ])

    def test_concat_placeholder(self):

        @dsl.container_component
        def comp(x: Optional[str] = None):
            return dsl.ContainerSpec(
                image='alpine',
                command=[dsl.ConcatPlaceholder(['prefix-', x, '-suffix'])])

        comp()

        run_mock = self.mocked_docker_client.containers.run
        run_mock.assert_called_once()
        kwargs = run_mock.call_args[1]
        self.assertEqual(
            kwargs['image'],
            'alpine:latest',
        )
        self.assertEqual(kwargs['command'], ['prefix-null-suffix'])

    def test_nested_concat_placeholder(self):

        @dsl.container_component
        def comp(x: Optional[str] = None):
            return dsl.ContainerSpec(
                image='alpine',
                command=[
                    'echo',
                    dsl.ConcatPlaceholder(
                        ['a', dsl.ConcatPlaceholder(['b', x, 'd'])])
                ])

        comp(x='c')

        run_mock = self.mocked_docker_client.containers.run
        run_mock.assert_called_once()
        kwargs = run_mock.call_args[1]
        self.assertEqual(
            kwargs['image'],
            'alpine:latest',
        )
        self.assertEqual(kwargs['command'], ['echo', 'abcd'])

    def test_ifpresent_in_concat_provided(self):

        @dsl.container_component
        def comp(x: Optional[str] = None):
            return dsl.ContainerSpec(
                image='alpine',
                command=[
                    'echo',
                    dsl.ConcatPlaceholder([
                        'there ',
                        dsl.ConcatPlaceholder([
                            'is ',
                            dsl.IfPresentPlaceholder(
                                input_name='x',
                                then='one thing',
                                else_='another thing')
                        ])
                    ])
                ])

        comp(x='c')

        run_mock = self.mocked_docker_client.containers.run
        run_mock.assert_called_once()
        kwargs = run_mock.call_args[1]
        self.assertEqual(
            kwargs['image'],
            'alpine:latest',
        )
        self.assertEqual(kwargs['command'], ['echo', 'there is one thing'])

    def test_ifpresent_in_concat_omitted(self):

        @dsl.container_component
        def comp(x: Optional[str] = None):
            return dsl.ContainerSpec(
                image='alpine',
                command=[
                    'echo',
                    dsl.ConcatPlaceholder([
                        'there ',
                        dsl.ConcatPlaceholder([
                            'is ',
                            dsl.IfPresentPlaceholder(
                                input_name='x',
                                then='one thing',
                                else_='another thing')
                        ])
                    ])
                ])

        comp()

        run_mock = self.mocked_docker_client.containers.run
        run_mock.assert_called_once()
        kwargs = run_mock.call_args[1]
        self.assertEqual(
            kwargs['image'],
            'alpine:latest',
        )
        self.assertEqual(kwargs['command'], ['echo', 'there is another thing'])

    def test_concat_in_ifpresent_provided(self):

        @dsl.container_component
        def comp(x: Optional[str] = None):
            return dsl.ContainerSpec(
                image='alpine',
                command=[
                    'echo',
                    dsl.IfPresentPlaceholder(
                        input_name='x',
                        then=dsl.ConcatPlaceholder([x]),
                        else_=dsl.ConcatPlaceholder(['something', ' ', 'else']))
                ])

        comp(x='something')

        run_mock = self.mocked_docker_client.containers.run
        run_mock.assert_called_once()
        kwargs = run_mock.call_args[1]
        self.assertEqual(
            kwargs['image'],
            'alpine:latest',
        )
        self.assertEqual(kwargs['command'], ['echo', 'something'])

    def test_concat_in_ifpresent_omitted(self):

        @dsl.container_component
        def comp(x: Optional[str] = None):
            return dsl.ContainerSpec(
                image='alpine',
                command=[
                    'echo',
                    dsl.IfPresentPlaceholder(
                        input_name='x',
                        then=dsl.ConcatPlaceholder([x]),
                        else_=dsl.ConcatPlaceholder(['another', ' ', 'thing']))
                ])

        comp()

        run_mock = self.mocked_docker_client.containers.run
        run_mock.assert_called_once()
        kwargs = run_mock.call_args[1]
        self.assertEqual(
            kwargs['image'],
            'alpine:latest',
        )
        self.assertEqual(kwargs['command'], ['echo', 'another thing'])


if __name__ == '__main__':
    unittest.main()



# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility methods for compiler implementation that is IR-agnostic."""

import collections
import copy

# isort: off
# Disable isort because isort and yapf conflict with each other.
from typing import (Any, DefaultDict, Dict, List, Mapping, Optional, Set, Tuple,
                    Union)
# isort: on

from kfp import dsl
from kfp.dsl import constants
from kfp.dsl import for_loop
from kfp.dsl import pipeline_channel
from kfp.dsl import pipeline_context
from kfp.dsl import pipeline_task
from kfp.dsl import tasks_group

GroupOrTaskType = Union[tasks_group.TasksGroup, pipeline_task.PipelineTask]

ILLEGAL_CROSS_DAG_ERROR_PREFIX = 'Illegal task dependency across DSL context managers.'


def additional_input_name_for_pipeline_channel(
        channel_or_name: Union[pipeline_channel.PipelineChannel, str]) -> str:
    """Gets the name for an additional (compiler-injected) input."""

    # Adding a prefix to avoid (reduce chance of) name collision between the
    # original component inputs and the injected input.
    return 'pipelinechannel--' + (
        channel_or_name.full_name if isinstance(
            channel_or_name, pipeline_channel.PipelineChannel) else
        channel_or_name)


def get_all_groups(
    root_group: tasks_group.TasksGroup,) -> List[tasks_group.TasksGroup]:
    """Gets all groups (not including tasks) in a pipeline.

    Args:
        root_group: The root group of a pipeline.

    Returns:
        A list of all groups in topological order (parent first).
    """
    all_groups = []

    def _get_all_groups_helper(
        group: tasks_group.TasksGroup,
        all_groups: List[tasks_group.TasksGroup],
    ):
        all_groups.append(group)
        for group in group.groups:
            _get_all_groups_helper(group, all_groups)

    _get_all_groups_helper(root_group, all_groups)
    return all_groups


def get_parent_groups(
    root_group: tasks_group.TasksGroup,
) -> Tuple[Mapping[str, List[str]], Mapping[str, List[str]]]:
    """Get parent groups that contain the specified tasks.

    Each pipeline has a root group. Each group has a list of tasks (leaf)
    and groups.
    This function traverse the tree and get ancestor groups for all tasks.

    Args:
        root_group: The root group of a pipeline.

    Returns:
        A tuple. The first item is a mapping of task names to parent groups,
        and second item is a mapping of group names to parent groups.
        A list of parent groups is a list of ancestor groups including the
        task/group itself. The list is sorted in a way that the farthest
        parent group is the first and task/group itself is the last.
    """

    def _get_parent_groups_helper(
        current_groups: List[tasks_group.TasksGroup],
        tasks_to_groups: Dict[str, List[GroupOrTaskType]],
        groups_to_groups: Dict[str, List[GroupOrTaskType]],
    ) -> None:
        root_group = current_groups[-1]
        for group in root_group.groups:

            groups_to_groups[group.name] = [x.name for x in current_groups
                                           ] + [group.name]
            current_groups.append(group)

            _get_parent_groups_helper(
                current_groups=current_groups,
                tasks_to_groups=tasks_to_groups,
                groups_to_groups=groups_to_groups,
            )
            del current_groups[-1]

        for task in root_group.tasks:
            tasks_to_groups[task.name] = [x.name for x in current_groups
                                         ] + [task.name]

    tasks_to_groups = {}
    groups_to_groups = {}
    current_groups = [root_group]

    _get_parent_groups_helper(
        current_groups=current_groups,
        tasks_to_groups=tasks_to_groups,
        groups_to_groups=groups_to_groups,
    )
    return (tasks_to_groups, groups_to_groups)


def get_channels_from_condition(
    operations: List[pipeline_channel.ConditionOperation],
    collected_channels: list,
) -> None:
    """Appends to collected_channels each pipeline channels used in each
    operand of each operation in operations."""
    for operation in operations:
        for operand in [operation.left_operand, operation.right_operand]:
            if isinstance(operand, pipeline_channel.PipelineChannel):
                collected_channels.append(operand)


def get_condition_channels_for_tasks(
    root_group: tasks_group.TasksGroup,
) -> Mapping[str, Set[pipeline_channel.PipelineChannel]]:
    """Gets channels referenced in conditions of tasks' parents.

    Args:
        root_group: The root group of a pipeline.

    Returns:
        A mapping of task name to a set of pipeline channels appeared in its
        parent dsl.Condition groups.
    """
    conditions = collections.defaultdict(set)

    def _get_condition_channels_for_tasks_helper(
        group,
        current_conditions_channels,
    ):
        new_current_conditions_channels = current_conditions_channels
        if isinstance(group, tasks_group._ConditionBase):
            new_current_conditions_channels = list(current_conditions_channels)
            get_channels_from_condition(
                group.conditions,
                new_current_conditions_channels,
            )

        for task in group.tasks:
            for channel in new_current_conditions_channels:
                conditions[task.name].add(channel)
        for group in group.groups:
            _get_condition_channels_for_tasks_helper(
                group, new_current_conditions_channels)

    _get_condition_channels_for_tasks_helper(root_group, [])
    return conditions


def get_inputs_for_all_groups(
    pipeline: pipeline_context.Pipeline,
    task_name_to_parent_groups: Mapping[str, List[str]],
    group_name_to_parent_groups: Mapping[str, List[str]],
    condition_channels: Mapping[str,
                                Set[pipeline_channel.PipelineParameterChannel]],
    name_to_for_loop_group: Mapping[str, tasks_group.ParallelFor],
) -> Mapping[str, List[Tuple[pipeline_channel.PipelineChannel, str]]]:
    """Get inputs and outputs of each group and op.

    Args:
        pipeline: The instantiated pipeline object.
        task_name_to_parent_groups: The dict of task name to list of parent
            groups.
        group_name_to_parent_groups: The dict of group name to list of
            parent groups.
        condition_channels: The dict of task name to a set of pipeline
            channels referenced by its parent condition groups.
        name_to_for_loop_group: The dict of for loop group name to loop
            group.

    Returns:
        A mapping  with key being the group/task names and values being list
        of tuples (channel, producing_task_name).
        producing_task_name is the name of the task that produces the
        channel. If the channel is a pipeline argument (no producer task),
        then producing_task_name is None.
    """
    inputs = collections.defaultdict(set)

    for task in pipeline.tasks.values():
        # task's inputs and all channels used in conditions for that task are
        # considered.
        task_condition_inputs = list(condition_channels[task.name])

        for channel in task.channel_inputs + task_condition_inputs:

            # If the value is already provided (immediate value), then no
            # need to expose it as input for its parent groups.
            if getattr(channel, 'value', None):
                continue

            # channels_to_add could be a list of PipelineChannels when loop
            # args are involved. Given a nested loops example as follows:
            #
            #  def my_pipeline(loop_parameter: list):
            #       with dsl.ParallelFor(loop_parameter) as item:
            #           with dsl.ParallelFor(item.p_a) as item_p_a:
            #               print_op(item_p_a.q_a)
            #
            # The print_op takes an input of
            # {{channel:task=;name=loop_parameter-loop-item-subvar-p_a-loop-item-subvar-q_a;}}.
            # Given this, we calculate the list of PipelineChannels potentially
            # needed by across DAG levels as follows:
            #
            # [{{channel:task=;name=loop_parameter-loop-item-subvar-p_a-loop-item-subvar-q_a}},
            #  {{channel:task=;name=loop_parameter-loop-item-subvar-p_a-loop-item}},
            #  {{channel:task=;name=loop_parameter-loop-item-subvar-p_a}},
            #  {{channel:task=;name=loop_parameter-loop-item}},
            #  {{chaenel:task=;name=loop_parameter}}]
            #
            # For the above example, the first loop needs the input of
            # {{channel:task=;name=loop_parameter}},
            # the second loop needs the input of
            # {{channel:task=;name=loop_parameter-loop-item}}
            # and the print_op needs the input of
            # {{channel:task=;name=loop_parameter-loop-item-subvar-p_a-loop-item}}
            #
            # When we traverse a DAG in a top-down direction, we add channels
            # from the end, and pop it out when it's no longer needed by the
            # sub-DAG.
            # When we traverse a DAG in a bottom-up direction, we add
            # channels from the front, and pop it out when it's no longer
            #  needed by the parent DAG.
            channels_to_add = collections.deque()
            channel_to_add = channel

            while isinstance(channel_to_add, (
                    for_loop.LoopParameterArgument,
                    for_loop.LoopArtifactArgument,
                    for_loop.LoopArgumentVariable,
            )):
                channels_to_add.append(channel_to_add)
                if isinstance(channel_to_add, for_loop.LoopArgumentVariable):
                    channel_to_add = channel_to_add.loop_argument
                else:
                    channel_to_add = channel_to_add.items_or_pipeline_channel

            if isinstance(channel_to_add, pipeline_channel.PipelineChannel):
                channels_to_add.append(channel_to_add)

            if channel.task:
                # The PipelineChannel is produced by a task.
                upstream_task = channel.task
                upstream_groups, downstream_groups = (
                    _get_uncommon_ancestors(
                        task_name_to_parent_groups=task_name_to_parent_groups,
                        group_name_to_parent_groups=group_name_to_parent_groups,
                        task1=upstream_task,
                        task2=task,
                    ))

                for i, group_name in enumerate(downstream_groups):
                    if i == 0:
                        # If it is the first uncommon downstream group, then
                        # the input comes from the first uncommon upstream
                        # group.
                        producer_task = upstream_groups[0]
                    else:
                        # If not the first downstream group, then the input
                        # is passed down from its ancestor groups so the
                        # upstream group is None.
                        producer_task = None

                    inputs[group_name].add((channels_to_add[-1], producer_task))

                    if group_name in name_to_for_loop_group:
                        loop_group = name_to_for_loop_group[group_name]

                        # Pop out the last elements from channels_to_add if it
                        # is found in the current (loop) DAG. Downstreams
                        # would only need the more specific versions for it.
                        if channels_to_add[
                                -1].full_name in loop_group.loop_argument.full_name:
                            channels_to_add.pop()
                            if not channels_to_add:
                                break

            else:
                # The PipelineChannel is not produced by a task. It's either
                # a top-level pipeline input, or a constant value to loop
                # items.

                # TODO: revisit if this is correct.
                if getattr(task, 'is_exit_handler', False):
                    continue

                # For PipelineChannel as a result of constant value used as
                # loop items, we have to go from bottom-up because the
                # PipelineChannel can be originated from the middle a DAG,
                # which is not needed and visible to its parent DAG.
                if isinstance(channel, (
                        for_loop.LoopParameterArgument,
                        for_loop.LoopArtifactArgument,
                        for_loop.LoopArgumentVariable,
                )) and channel.is_with_items_loop_argument:
                    for group_name in task_name_to_parent_groups[
                            task.name][::-1]:

                        inputs[group_name].add((channels_to_add[0], None))
                        if group_name in name_to_for_loop_group:
                            # for example:
                            #   loop_group.loop_argument.name = 'loop-item-param-1'
                            #   channel.name = 'loop-item-param-1-subvar-a'
                            loop_group = name_to_for_loop_group[group_name]

                            if channels_to_add[
                                    0].full_name in loop_group.loop_argument.full_name:
                                channels_to_add.popleft()
                                if not channels_to_add:
                                    break
                else:
                    # For PipelineChannel from pipeline input, go top-down
                    # just like we do for PipelineChannel produced by a task.
                    for group_name in task_name_to_parent_groups[task.name]:

                        inputs[group_name].add((channels_to_add[-1], None))
                        if group_name in name_to_for_loop_group:
                            loop_group = name_to_for_loop_group[group_name]

                            if channels_to_add[
                                    -1].full_name in loop_group.loop_argument.full_name:
                                channels_to_add.pop()
                                if not channels_to_add:
                                    break

    return inputs


class InvalidTopologyException(Exception):
    pass


def validate_parallel_for_fan_in_consumption_legal(
    consumer_task_name: str,
    upstream_groups: List[str],
    group_name_to_group: Dict[str, tasks_group.TasksGroup],
) -> None:
    """Checks that a dsl.Collected object is being used results in an
    unambiguous pipeline topology and is therefore legal.

    Args:
        consumer_task_name: The name of the consumer task.
        upstream_groups: The names of the producer task's upstream groups, ordered from outermost group at beginning to producer task at end. This is produced by produced by _get_uncommon_ancestors.
        group_name_to_group: Map of group name to TasksGroup, for fast lookups.
    """
    # handles cases like this:
    # @dsl.pipeline
    # def my_pipeline():
    #     with dsl.ParallelFor([1, 2, 3]) as x:
    #         t = double(num=x)
    #         x = add(dsl.Collected(t.output))
    #
    # and this:
    # @dsl.pipeline
    # def my_pipeline():
    #     t = double(num=1)
    #     x = add(dsl.Collected(t.output))
    producer_task_idx = -1
    producer_task_name = upstream_groups[producer_task_idx]
    if all(group_name_to_group[group_name].group_type !=
           tasks_group.TasksGroupType.FOR_LOOP
           for group_name in upstream_groups[:producer_task_idx]):
        raise InvalidTopologyException(
            f'dsl.{for_loop.Collected.__name__} can only be used to fan-in outputs produced by a task within a dsl.{tasks_group.ParallelFor.__name__} context to a task outside of the dsl.{tasks_group.ParallelFor.__name__} context. Producer task {producer_task_name} is either not in a dsl.{tasks_group.ParallelFor.__name__} context or is only in a dsl.{tasks_group.ParallelFor.__name__} that also contains consumer task {consumer_task_name}.'
        )

    # illegal if the producer has a parent conditional outside of its outermost for loop, since the for loop may or may not be executed
    # for example, what happens if text == 'b'? the resulting execution behavior is ambiguous.
    #
    # @dsl.pipeline
    # def my_pipeline(text: str = ''):
    #     with dsl.Condition(text == 'a'):
    #         with dsl.ParallelFor([1, 2, 3]) as x:
    #             t = double(num=x)
    #     x = add(nums=dsl.Collected(t.output))
    outermost_uncommon_upstream_group = upstream_groups[0]
    group = group_name_to_group[outermost_uncommon_upstream_group]
    if group.group_type in [
            tasks_group.TasksGroupType.CONDITION,
            tasks_group.TasksGroupType.EXIT_HANDLER,
    ]:
        raise InvalidTopologyException(
            f'{ILLEGAL_CROSS_DAG_ERROR_PREFIX} When using dsl.{for_loop.Collected.__name__} to fan-in outputs from a task within a dsl.{tasks_group.ParallelFor.__name__} context, the dsl.{tasks_group.ParallelFor.__name__} context manager cannot be nested within a dsl.{group.__class__.__name__} context manager unless the consumer task is too. Task {consumer_task_name} consumes from {producer_task_name} within a dsl.{group.__class__.__name__} context.'
        )
    elif group.group_type != tasks_group.TasksGroupType.FOR_LOOP:
        raise ValueError(
            f'Got unexpected group type when validating fanning-in outputs from task in dsl.{tasks_group.ParallelFor.__name__}: {group.group_type}'
        )


def make_new_channel_for_collected_outputs(
    channel_name: str,
    starting_channel: pipeline_channel.PipelineChannel,
    task_name: str,
) -> pipeline_channel.PipelineChannel:
    """Creates a new PipelineParameterChannel/PipelineArtifactChannel list for
    a dsl.Collected channel from the original task output."""
    if isinstance(starting_channel, pipeline_channel.PipelineParameterChannel):
        return pipeline_channel.PipelineParameterChannel(
            channel_name, channel_type='LIST', task_name=task_name)
    elif isinstance(starting_channel, pipeline_channel.PipelineArtifactChannel):
        return pipeline_channel.PipelineArtifactChannel(
            channel_name,
            channel_type=starting_channel.channel_type,
            task_name=task_name,
            is_artifact_list=True)
    else:
        ValueError(
            f'Got unknown PipelineChannel: {starting_channel!r}. Expected an instance of {pipeline_channel.PipelineArtifactChannel.__name__!r} or {pipeline_channel.PipelineParameterChannel.__name__!r}.'
        )


def get_outputs_for_all_groups(
    pipeline: pipeline_context.Pipeline,
    task_name_to_parent_groups: Mapping[str, List[str]],
    group_name_to_parent_groups: Mapping[str, List[str]],
    all_groups: List[tasks_group.TasksGroup],
    pipeline_outputs_dict: Dict[str, pipeline_channel.PipelineChannel]
) -> Tuple[DefaultDict[str, Dict[str, pipeline_channel.PipelineChannel]], Dict[
        str, pipeline_channel.PipelineChannel]]:
    """Gets a dictionary of all TasksGroup names to an inner dictionary. The
    inner dictionary is TasksGroup output keys to channels corresponding to
    those keys.

    It constructs this dictionary from both data passing within the pipeline body, as well as the outputs returned from the pipeline (e.g., return dsl.Collected(...)).

    Also returns as the second item of tuple the updated pipeline_outputs_dict. This dict is modified so that the values (PipelineChannel) references the group that surfaces the task output, instead of the original task that produced it.
    """

    # unlike inputs, which will be surfaced as component input parameters,
    # consumers of surfaced outputs need to have a reference to what the parent
    # component calls them when they surface them, which will be different than
    # the producer task name and channel name (the information contained in the
    # pipeline channel)
    # for this reason, we use additional_input_name_for_pipeline_channel here
    # to set the name of the surfaced output once

    group_name_to_group = {group.name: group for group in all_groups}
    group_name_to_children = {
        group.name: [group.name for group in group.groups] +
        [task.name for task in group.tasks] for group in all_groups
    }

    outputs = collections.defaultdict(dict)
    processed_oneofs: Set[pipeline_channel.OneOfMixin] = set()
    # handle dsl.Collected consumed by tasks
    for task in pipeline.tasks.values():
        for channel in task.channel_inputs:
            # TODO: migrate Collected to OneOfMixin style implementation,
            # then simplify this logic to align with OneOfMixin logic
            if isinstance(channel, dsl.Collected):
                producer_task = pipeline.tasks[channel.task_name]
                consumer_task = task

                upstream_groups, downstream_groups = (
                    _get_uncommon_ancestors(
                        task_name_to_parent_groups=task_name_to_parent_groups,
                        group_name_to_parent_groups=group_name_to_parent_groups,
                        task1=producer_task,
                        task2=consumer_task,
                    ))
                validate_parallel_for_fan_in_consumption_legal(
                    consumer_task_name=consumer_task.name,
                    upstream_groups=upstream_groups,
                    group_name_to_group=group_name_to_group,
                )

                # producer_task's immediate parent group and the name by which
                # to surface the channel
                surfaced_output_name = additional_input_name_for_pipeline_channel(
                    channel)

                # the highest-level task group that "consumes" the
                # collected output
                parent_consumer = downstream_groups[0]
                producer_task_name = upstream_groups.pop()

                # process from the upstream groups from the inside out
                for upstream_name in reversed(upstream_groups):
                    outputs[upstream_name][
                        surfaced_output_name] = make_new_channel_for_collected_outputs(
                            channel_name=channel.name,
                            starting_channel=channel.output,
                            task_name=producer_task_name,
                        )

                    # on each iteration, mutate the channel being consumed so
                    # that it references the last parent group surfacer
                    channel.name = surfaced_output_name
                    channel.task_name = upstream_name

                    # for the next iteration, set the consumer to the current
                    # surfacer (parent group)
                    producer_task_name = upstream_name

                    parent_of_current_surfacer = group_name_to_parent_groups[
                        upstream_name][-2]
                    if parent_consumer in group_name_to_children[
                            parent_of_current_surfacer]:
                        break

            elif isinstance(channel, pipeline_channel.OneOfMixin):
                if channel in processed_oneofs:
                    continue

                # we want to mutate the oneof's inner channels ONLY where they
                # are used in the oneof, not if they are used separately
                # for example: we should only modify the copy of
                # foo.output in dsl.OneOf(foo.output), not if foo.output is
                # passed to another downstream task
                channel.channels = [copy.copy(c) for c in channel.channels]
                for inner_channel in channel.channels:
                    producer_task = pipeline.tasks[inner_channel.task_name]
                    consumer_task = task
                    upstream_groups, downstream_groups = (
                        _get_uncommon_ancestors(
                            task_name_to_parent_groups=task_name_to_parent_groups,
                            group_name_to_parent_groups=group_name_to_parent_groups,
                            task1=producer_task,
                            task2=consumer_task,
                        ))
                    surfaced_output_name = additional_input_name_for_pipeline_channel(
                        inner_channel)

                    # 1. get the oneof
                    # 2. find the task group that surfaced it
                    # 3. find the inner tasks reponsible

                    for upstream_name in reversed(upstream_groups):
                        # skip the first task processed, since we don't need to add new outputs for the innermost task
                        if upstream_name == inner_channel.task.name:
                            continue
                        # # once we've hit the outermost condition-branches group, we're done
                        if upstream_name == channel.condition_branches_group.name:
                            outputs[upstream_name][channel.name] = channel
                            break

                        # copy as a mechanism for "freezing" the inner channel
                        # before we make updates for the next iteration
                        outputs[upstream_name][
                            surfaced_output_name] = copy.copy(inner_channel)

                        inner_channel.name = surfaced_output_name
                        inner_channel.task_name = upstream_name

                processed_oneofs.add(channel)

    # handle dsl.Collected returned from pipeline
    # TODO: consider migrating dsl.Collected returns to pattern used by dsl.OneOf, where the OneOf constructor returns a parameter/artifact channel, which fits in more cleanly into the existing compiler abtractions
    for output_key, channel in pipeline_outputs_dict.items():
        if isinstance(channel, for_loop.Collected):
            surfaced_output_name = additional_input_name_for_pipeline_channel(
                channel)
            upstream_groups = task_name_to_parent_groups[channel.task_name][1:]
            producer_task_name = upstream_groups.pop()
            # process upstream groups from the inside out, until getting to the pipeline level
            for upstream_name in reversed(upstream_groups):
                new_channel = make_new_channel_for_collected_outputs(
                    channel_name=channel.name,
                    starting_channel=channel.output,
                    task_name=producer_task_name,
                )

                # on each iteration, mutate the channel being consumed so
                # that it references the last parent group surfacer
                channel.name = surfaced_output_name
                channel.task_name = upstream_name

                # for the next iteration, set the consumer to the current
                # surfacer (parent group)
                producer_task_name = upstream_name
                outputs[upstream_name][surfaced_output_name] = new_channel

            # after surfacing from all inner TasksGroup, change the PipelineChannel output to also return from the correct TasksGroup
            pipeline_outputs_dict[
                output_key] = make_new_channel_for_collected_outputs(
                    channel_name=surfaced_output_name,
                    starting_channel=channel.output,
                    task_name=upstream_name,
                )
        elif isinstance(channel, pipeline_channel.OneOfMixin):
            # if the output has already been consumed by a task before it is returned, we don't need to reprocess it
            if channel in processed_oneofs:
                continue

            # we want to mutate the oneof's inner channels ONLY where they
            # are used in the oneof, not if they are used separately
            # for example: we should only modify the copy of
            # foo.output in dsl.OneOf(foo.output), not if foo.output is passed
            # to another downstream task
            channel.channels = [copy.copy(c) for c in channel.channels]
            for inner_channel in channel.channels:
                producer_task = pipeline.tasks[inner_channel.task_name]
                upstream_groups = task_name_to_parent_groups[
                    inner_channel.task_name][1:]
                surfaced_output_name = additional_input_name_for_pipeline_channel(
                    inner_channel)

                # 1. get the oneof
                # 2. find the task group that surfaced it
                # 3. find the inner tasks reponsible
                for upstream_name in reversed(upstream_groups):
                    # skip the first task processed, since we don't need to add new outputs for the innermost task
                    if upstream_name == inner_channel.task.name:
                        continue
                    # # once we've hit the outermost condition-branches group, we're done
                    if upstream_name == channel.condition_branches_group.name:
                        outputs[upstream_name][channel.name] = channel
                        break

                    # copy as a mechanism for "freezing" the inner channel
                    # before we make updates for the next iteration
                    outputs[upstream_name][surfaced_output_name] = copy.copy(
                        inner_channel)

                    inner_channel.name = surfaced_output_name
                    inner_channel.task_name = upstream_name
    return outputs, pipeline_outputs_dict


def _get_uncommon_ancestors(
    task_name_to_parent_groups: Mapping[str, List[str]],
    group_name_to_parent_groups: Mapping[str, List[str]],
    task1: GroupOrTaskType,
    task2: GroupOrTaskType,
) -> Tuple[List[str], List[str]]:
    """Gets the unique ancestors between two tasks.

    For example, task1's ancestor groups are [root, G1, G2, G3, task1],
    task2's ancestor groups are [root, G1, G4, task2], then it returns a
    tuple ([G2, G3, task1], [G4, task2]).

    Args:
        task_name_to_parent_groups: The dict of task name to list of parent
            groups.
        group_name_tor_parent_groups: The dict of group name to list of
            parent groups.
        task1: One of the two tasks.
        task2: The other task.

    Returns:
        A tuple which are lists of uncommon ancestors for each task.
    """
    if task1.name in task_name_to_parent_groups:
        task1_groups = task_name_to_parent_groups[task1.name]
    elif task1.name in group_name_to_parent_groups:
        task1_groups = group_name_to_parent_groups[task1.name]
    else:
        raise ValueError(task1.name + ' does not exist.')

    if task2.name in task_name_to_parent_groups:
        task2_groups = task_name_to_parent_groups[task2.name]
    elif task2.name in group_name_to_parent_groups:
        task2_groups = group_name_to_parent_groups[task2.name]
    else:
        raise ValueError(task2.name + ' does not exist.')

    both_groups = [task1_groups, task2_groups]
    common_groups_len = sum(
        1 for x in zip(*both_groups) if x == (x[0],) * len(x))
    group1 = task1_groups[common_groups_len:]
    group2 = task2_groups[common_groups_len:]
    return (group1, group2)


def get_dependencies(
    pipeline: pipeline_context.Pipeline,
    task_name_to_parent_groups: Mapping[str, List[str]],
    group_name_to_parent_groups: Mapping[str, List[str]],
    group_name_to_group: Mapping[str, tasks_group.TasksGroup],
    condition_channels: Dict[str, pipeline_channel.PipelineChannel],
) -> Mapping[str, List[GroupOrTaskType]]:
    """Gets dependent groups and tasks for all tasks and groups.

    Args:
        pipeline: The instantiated pipeline object.
        task_name_to_parent_groups: The dict of task name to list of parent
            groups.
        group_name_to_parent_groups: The dict of group name to list of
            parent groups.
        group_name_to_group: The dict of group name to group.
        condition_channels: The dict of task name to a set of pipeline
            channels referenced by its parent condition groups.

    Returns:
        A Mapping where key is group/task name, value is a list of dependent
        groups/tasks. The dependencies are calculated in the following way:
        if task2 depends on task1, and their ancestors are
        [root, G1, G2, task1] and [root, G1, G3, G4, task2], then G3 is
        dependent on G2. Basically dependency only exists in the first
        uncommon ancesters in their ancesters chain. Only sibling
        groups/tasks can have dependencies.

    Raises:
        RuntimeError: if a task depends on a task inside a condition or loop
            group.
    """
    dependencies = collections.defaultdict(set)
    for task in pipeline.tasks.values():
        upstream_task_names: Set[Union[pipeline_task.PipelineTask,
                                       tasks_group.TasksGroup]] = set()
        task_condition_inputs = list(condition_channels[task.name])
        all_channels = task.channel_inputs + task_condition_inputs
        upstream_task_names.update(
            {channel.task for channel in all_channels if channel.task})
        # dependent tasks is tasks on which .after was called and can only be the names of PipelineTasks, not TasksGroups
        upstream_task_names.update(
            {pipeline.tasks[after_task] for after_task in task.dependent_tasks})

        for upstream_task in upstream_task_names:

            upstream_names, downstream_names = _get_uncommon_ancestors(
                task_name_to_parent_groups=task_name_to_parent_groups,
                group_name_to_parent_groups=group_name_to_parent_groups,
                task1=upstream_task,
                task2=task,
            )

            # uncommon upstream ancestor check
            uncommon_upstream_names = copy.deepcopy(upstream_names)
            # because a task's `upstream_groups` contains the upstream task's name, remove it so we can check it's parent TasksGroups
            uncommon_upstream_names.remove(upstream_task.name)

            if uncommon_upstream_names:
                upstream_parent_group = group_name_to_group.get(
                    uncommon_upstream_names[0], None)

                # downstream tasks cannot depend on tasks in a dsl.ExitHandler or condition context if the downstream task is not also in the same context
                # this applies for .after() and data exchange
                if isinstance(
                        upstream_parent_group,
                    (tasks_group._ConditionBase, tasks_group.ExitHandler)):
                    raise InvalidTopologyException(
                        f'{ILLEGAL_CROSS_DAG_ERROR_PREFIX} A downstream task cannot depend on an upstream task within a dsl.{upstream_parent_group.__class__.__name__} context unless the downstream is within that context too. Found task {task.name} which depends on upstream task {upstream_task.name} within an uncommon dsl.{upstream_parent_group.__class__.__name__} context.'
                    )
                # same for dsl.ParallelFor, but only throw on data exchange (.after is allowed)
                # TODO: migrate Collected to OneOfMixin style implementation,
                # then make this validation dsl.Collected-aware
                elif isinstance(upstream_parent_group, tasks_group.ParallelFor):
                    upstream_tasks_that_downstream_consumers_from = [
                        channel.task.name
                        for channel in task._channel_inputs
                        if channel.task is not None
                    ]
                    has_data_exchange = upstream_task.name in upstream_tasks_that_downstream_consumers_from
                    # don't raise for .after
                    if has_data_exchange:
                        raise InvalidTopologyException(
                            f'{ILLEGAL_CROSS_DAG_ERROR_PREFIX} A downstream task cannot depend on an upstream task within a dsl.{upstream_parent_group.__class__.__name__} context unless the downstream is within that context too or the outputs are begin fanned-in to a list using dsl.{for_loop.Collected.__name__}. Found task {task.name} which depends on upstream task {upstream_task.name} within an uncommon dsl.{upstream_parent_group.__class__.__name__} context.'
                        )

            dependencies[downstream_names[0]].add(upstream_names[0])

    return dependencies


def recursive_replace_placeholders(data: Any, old_value: str,
                                   new_value: str) -> Union[Dict, List, str]:
    """Replaces the given data. If the data is a dict/list object, it
    recursively replaces values in it.

    This method is used to replace PipelineChannel objects with input parameter
    placeholders in a nested object like worker_pool_specs for custom jobs.

    Args:
        data: An object, which could be a nested object including dictionaries and lists.
        old_value: The value that will be replaced.
        new_value: The value to replace the old value with.

    Returns:
        A copy of data with all occurences of old_value replaced by new_value.
    """
    if isinstance(data, dict):
        return {
            k: recursive_replace_placeholders(v, old_value, new_value)
            for k, v in data.items()
        }
    elif isinstance(data, list):
        return [
            recursive_replace_placeholders(i, old_value, new_value)
            for i in data
        ]
    else:
        if isinstance(data, pipeline_channel.PipelineChannel):
            data = str(data)
        if isinstance(data, str):
            return data.replace(old_value, new_value)
        else:
            return data


# Note that cpu_to_float assumes the string has already been validated by the _validate_cpu_request_limit method.
def _cpu_to_float(cpu: str) -> float:
    """Converts the validated CPU request/limit string and to its numeric float
    value.

    Args:
        cpu: CPU requests or limits. This string should be a number or a
            number followed by an "m" to indicate millicores (1/1000). For
            more information, see `Specify a CPU Request and a CPU Limit
    Returns:
        The numeric value (float) of the cpu request/limit.
    """
    return float(cpu[:-1]) / 1000 if cpu.endswith('m') else float(cpu)


# Note that memory_to_float assumes the string has already been validated by the _validate_memory_request_limit method.
def _memory_to_float(memory: str) -> float:
    """Converts the validated memory request/limit string to its numeric value.

    Args:
        memory: Memory requests or limits. This string should be a number or
            a number followed by one of "E", "Ei", "P", "Pi", "T", "Ti", "G",
            "Gi", "M", "Mi", "K", or "Ki".
    Returns:
        The numeric value (float) of the memory request/limit.
    """
    if memory.endswith('E'):
        memory = float(memory[:-1]) * constants._E / constants._G
    elif memory.endswith('Ei'):
        memory = float(memory[:-2]) * constants._EI / constants._G
    elif memory.endswith('P'):
        memory = float(memory[:-1]) * constants._P / constants._G
    elif memory.endswith('Pi'):
        memory = float(memory[:-2]) * constants._PI / constants._G
    elif memory.endswith('T'):
        memory = float(memory[:-1]) * constants._T / constants._G
    elif memory.endswith('Ti'):
        memory = float(memory[:-2]) * constants._TI / constants._G
    elif memory.endswith('G'):
        memory = float(memory[:-1])
    elif memory.endswith('Gi'):
        memory = float(memory[:-2]) * constants._GI / constants._G
    elif memory.endswith('M'):
        memory = float(memory[:-1]) * constants._M / constants._G
    elif memory.endswith('Mi'):
        memory = float(memory[:-2]) * constants._MI / constants._G
    elif memory.endswith('K'):
        memory = float(memory[:-1]) * constants._K / constants._G
    elif memory.endswith('Ki'):
        memory = float(memory[:-2]) * constants._KI / constants._G
    else:
        # By default interpret as a plain integer, in the unit of Bytes.
        memory = float(memory) / constants._G

    return memory


class KubernetesManifestOptions:
    """Options for Kubernetes manifest output during pipeline compilation.

    Args:
        pipeline_name: Name for the Pipeline resource (defaults to pipeline_spec.pipeline_info.name).
        pipeline_display_name: Display name for the Pipeline resource (defaults to pipeline_name).
        pipeline_version_name: Name for the PipelineVersion resource (defaults to pipeline_name).
        pipeline_version_display_name: Display name for the PipelineVersion resource (defaults to pipeline_display_name).
        namespace: Kubernetes namespace for the resources (optional).
        include_pipeline_manifest: Whether to include the Pipeline manifest (default: False).
    """

    def __init__(
        self,
        pipeline_name: Optional[str] = None,
        pipeline_display_name: Optional[str] = None,
        pipeline_version_name: Optional[str] = None,
        pipeline_version_display_name: Optional[str] = None,
        namespace: Optional[str] = None,
        include_pipeline_manifest: bool = False,
    ):
        self.pipeline_name = pipeline_name
        self.pipeline_display_name = pipeline_display_name
        self.pipeline_version_name = pipeline_version_name
        self.pipeline_version_display_name = pipeline_version_display_name
        self.namespace = namespace
        self.include_pipeline_manifest = include_pipeline_manifest
        self._pipeline_spec = None

    def set_pipeline_spec(self, pipeline_spec):
        self._pipeline_spec = pipeline_spec

    @property
    def pipeline_name(self) -> Optional[str]:
        if self._pipeline_name is not None:
            return self._pipeline_name
        if self._pipeline_spec is not None:
            return self._pipeline_spec.pipeline_info.name
        return None

    @pipeline_name.setter
    def pipeline_name(self, value: Optional[str]):
        self._pipeline_name = value

    @property
    def pipeline_display_name(self) -> Optional[str]:
        return self._pipeline_display_name or self.pipeline_name

    @pipeline_display_name.setter
    def pipeline_display_name(self, value: Optional[str]):
        self._pipeline_display_name = value

    @property
    def pipeline_version_name(self) -> Optional[str]:
        return self._pipeline_version_name or self.pipeline_name

    @pipeline_version_name.setter
    def pipeline_version_name(self, value: Optional[str]):
        self._pipeline_version_name = value

    @property
    def pipeline_version_display_name(self) -> Optional[str]:
        return (self._pipeline_version_display_name or
                self.pipeline_version_name or self.pipeline_display_name)

    @pipeline_version_display_name.setter
    def pipeline_version_display_name(self, value: Optional[str]):
        self._pipeline_version_display_name = value



# Copyright 2024 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for graph_utils.py."""
from typing import Any, Dict
import unittest

from google.protobuf import json_format
from kfp.local import graph_utils
from kfp.pipeline_spec import pipeline_spec_pb2


class TestBuildDependencyMap(unittest.TestCase):

    def test_simple(self):
        tasks = {
            k: make_pipeline_task_spec(v)
            for k, v in SIMPLE_TASK_TOPOLOGY.items()
        }
        actual = graph_utils.build_dependency_map(tasks)
        expected = {'identity': [], 'identity-2': ['identity']}
        self.assertEqual(actual, expected)

    def test_complex(self):
        tasks = {
            k: make_pipeline_task_spec(v)
            for k, v in COMPLEX_TASK_TOPOLOGY.items()
        }
        actual = graph_utils.build_dependency_map(tasks)
        expected = {
            'add': [],
            'add-2': ['multiply'],
            'divide': ['add-2'],
            'multiply': ['add'],
            'printer': ['add', 'divide', 'multiply']
        }
        self.assertEqual(actual, expected)


class TestTopologicalSort(unittest.TestCase):

    def test_empty_graph(self):
        self.assertEqual(graph_utils.topological_sort({}), [])

    def test_simple_linear_graph(self):
        graph = {'A': ['B'], 'B': ['C'], 'C': []}
        actual = graph_utils.topological_sort(graph)
        expected = ['A', 'B', 'C']
        self.assertEqual(actual, expected)

    def test_separate_components(self):
        graph = {'A': ['B'], 'B': [], 'C': ['D'], 'D': []}
        actual = graph_utils.topological_sort(graph)
        expected = ['C', 'D', 'A', 'B']
        self.assertEqual(actual, expected)

    def test_complex_graph(self):
        graph = {'A': ['B', 'C'], 'B': ['D'], 'C': ['D'], 'D': []}
        actual = graph_utils.topological_sort(graph)
        expected = ['A', 'C', 'B', 'D']
        self.assertEqual(actual, expected)


class TestTopologicalSortTasks(unittest.TestCase):

    def test_simple(self):
        tasks = {
            k: make_pipeline_task_spec(v)
            for k, v in SIMPLE_TASK_TOPOLOGY.items()
        }
        actual = graph_utils.topological_sort_tasks(tasks)
        expected = ['identity-2', 'identity']
        self.assertEqual(actual, expected)

    def test_complex(self):
        tasks = {
            k: make_pipeline_task_spec(v)
            for k, v in COMPLEX_TASK_TOPOLOGY.items()
        }
        actual = graph_utils.topological_sort_tasks(tasks)
        expected = ['printer', 'divide', 'add-2', 'multiply', 'add']
        self.assertEqual(actual, expected)


SIMPLE_TASK_TOPOLOGY = {
    'identity': {},
    'identity-2': {
        'dependentTasks': ['identity'],
    }
}

COMPLEX_TASK_TOPOLOGY = {
    'add': {},
    'add-2': {
        'dependentTasks': ['multiply'],
    },
    'divide': {
        'dependentTasks': ['add-2'],
    },
    'multiply': {
        'dependentTasks': ['add'],
    },
    'printer': {
        'dependentTasks': ['add', 'divide', 'multiply'],
    }
}


def make_pipeline_task_spec(
        d: Dict[str, Any]) -> pipeline_spec_pb2.PipelineTaskSpec:
    spec = pipeline_spec_pb2.PipelineTaskSpec()
    json_format.ParseDict(d, spec)
    return spec


if __name__ == '__main__':
    unittest.main()



# Copyright 2023 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for reading and processing the ExecutorOutput message."""
import json
import os
from typing import Any, Dict, List, Union

from google.protobuf import json_format
from google.protobuf import struct_pb2
from kfp import dsl
from kfp.compiler import pipeline_spec_builder
from kfp.dsl import executor
from kfp.pipeline_spec import pipeline_spec_pb2


def load_executor_output(
        executor_output_path: str) -> pipeline_spec_pb2.ExecutorOutput:
    """Loads the ExecutorOutput message from a path."""
    executor_output = pipeline_spec_pb2.ExecutorOutput()

    if not os.path.isfile(executor_output_path):
        return executor_output

    with open(executor_output_path) as f:
        json_format.Parse(f.read(), executor_output)
    return executor_output


def cast_protobuf_numbers(
    output_parameters: Dict[str, Any],
    output_parameter_types: Dict[
        str, pipeline_spec_pb2.ComponentOutputsSpec.ParameterSpec],
) -> Dict[str, Any]:
    """Casts output fields that are typed as NUMBER_INTEGER to a Python int.

    This is required, since google.protobuf.struct_pb2.Value uses
    number_value to represent both floats and ints. When converting
    struct_pb2.Value to a dict/json, int will be upcast to float, even
    if the component output specifies int.
    """
    int_output_keys = [
        output_param_name
        for output_param_name, parameter_spec in output_parameter_types.items()
        if parameter_spec.parameter_type ==
        pipeline_spec_pb2.ParameterType.ParameterTypeEnum.NUMBER_INTEGER
    ]
    for int_output_key in int_output_keys:
        # avoid KeyError when the user never writes to the dsl.OutputPath
        if int_output_key in output_parameters:
            output_parameters[int_output_key] = int(
                output_parameters[int_output_key])
    return output_parameters


def get_outputs_from_executor_output(
    executor_output: pipeline_spec_pb2.ExecutorOutput,
    executor_input: pipeline_spec_pb2.ExecutorInput,
    component_spec: pipeline_spec_pb2.ComponentSpec,
) -> Dict[str, Any]:
    """Obtains a dictionary of output key to output value from several messages
    corresponding to the executed a component/task."""
    executor_output = add_type_to_executor_output(
        executor_input=executor_input,
        executor_output=executor_output,
    )

    # merge any parameter outputs written using dsl.OutputPath with the rest of ExecutorOutput
    executor_output = merge_dsl_output_file_parameters_to_executor_output(
        executor_input=executor_input,
        executor_output=executor_output,
        component_spec=component_spec,
    )

    # collect parameter outputs from executor output
    output_parameters = {
        param_name: pb2_value_to_python(value)
        for param_name, value in executor_output.parameter_values.items()
    }
    # process the special case of protobuf ints
    output_parameters = cast_protobuf_numbers(
        output_parameters,
        component_spec.output_definitions.parameters,
    )

    # collect artifact outputs from executor output
    output_artifact_definitions = component_spec.output_definitions.artifacts
    output_artifacts = {
        artifact_name: artifact_list_to_dsl_artifact(
            artifact_list,
            is_artifact_list=output_artifact_definitions[artifact_name]
            .is_artifact_list,
        ) for artifact_name, artifact_list in executor_output.artifacts.items()
    }
    return {**output_parameters, **output_artifacts}


def special_dsl_outputpath_read(
    parameter_name: str,
    output_file: str,
    dtype: pipeline_spec_pb2.ParameterType.ParameterTypeEnum,
) -> Any:
    """Reads the text in dsl.OutputPath files in the same way as the remote
    backend.

    In brief: read strings as strings and JSON load everything else.
    """
    try:
        with open(output_file) as f:
            value = f.read()

        if dtype == pipeline_spec_pb2.ParameterType.ParameterTypeEnum.STRING:
            value = value
        elif dtype == pipeline_spec_pb2.ParameterType.ParameterTypeEnum.BOOLEAN:
            # permit true/True and false/False, consistent with remote BE
            value = json.loads(value.lower())
        else:
            value = json.loads(value)
        return value
    except Exception as e:
        raise ValueError(
            f'Could not deserialize output {parameter_name!r} from path {output_file}'
        ) from e


def merge_dsl_output_file_parameters_to_executor_output(
    executor_input: pipeline_spec_pb2.ExecutorInput,
    executor_output: pipeline_spec_pb2.ExecutorOutput,
    component_spec: pipeline_spec_pb2.ComponentSpec,
) -> pipeline_spec_pb2.ExecutorOutput:
    """Merges and output parameters specified via dsl.OutputPath with the rest
    of the ExecutorOutput message."""
    for parameter_key, output_parameter in executor_input.outputs.parameters.items(
    ):
        if os.path.exists(output_parameter.output_file):
            parameter_value = special_dsl_outputpath_read(
                parameter_name=parameter_key,
                output_file=output_parameter.output_file,
                dtype=component_spec.output_definitions
                .parameters[parameter_key].parameter_type,
            )
            executor_output.parameter_values[parameter_key].CopyFrom(
                pipeline_spec_builder.to_protobuf_value(parameter_value))

    return executor_output


def pb2_value_to_python(value: struct_pb2.Value) -> Any:
    """Converts protobuf Value to the corresponding Python type."""
    if value.HasField('null_value'):
        return None
    elif value.HasField('number_value'):
        return value.number_value
    elif value.HasField('string_value'):
        return value.string_value
    elif value.HasField('bool_value'):
        return value.bool_value
    elif value.HasField('struct_value'):
        return pb2_struct_to_python(value.struct_value)
    elif value.HasField('list_value'):
        return [pb2_value_to_python(v) for v in value.list_value.values]
    else:
        raise ValueError(f'Unknown value type: {value}')


def pb2_struct_to_python(struct: struct_pb2.Struct) -> Dict[str, Any]:
    """Converts protobuf Struct to a dict."""
    return {k: pb2_value_to_python(v) for k, v in struct.fields.items()}


def runtime_artifact_to_dsl_artifact(
        runtime_artifact: pipeline_spec_pb2.RuntimeArtifact) -> dsl.Artifact:
    """Converts a single RuntimeArtifact instance to the corresponding
    dsl.Artifact instance."""
    return executor.create_artifact_instance(
        json_format.MessageToDict(runtime_artifact))


def artifact_list_to_dsl_artifact(
    artifact_list: pipeline_spec_pb2.ArtifactList,
    is_artifact_list: bool,
) -> Union[dsl.Artifact, List[dsl.Artifact]]:
    """Converts an ArtifactList instance to a single dsl.Artifact or a list of
    dsl.Artifacts, depending on whether the ArtifactList is a true list or
    simply a container for single artifact element."""
    dsl_artifacts = [
        runtime_artifact_to_dsl_artifact(artifact_spec)
        for artifact_spec in artifact_list.artifacts
    ]
    return dsl_artifacts if is_artifact_list else dsl_artifacts[0]


def add_type_to_executor_output(
    executor_input: pipeline_spec_pb2.ExecutorInput,
    executor_output: pipeline_spec_pb2.ExecutorOutput,
) -> pipeline_spec_pb2.ExecutorOutput:
    """Adds artifact type information (ArtifactTypeSchema) from the
    ExecutorInput message to the ExecutorOutput message.

    This information is not present in the ExecutorOutput message
    written by a task, though it is useful to have it for constructing
    local outputs. We don't want to change the executor logic and the
    serialized outputs of all tasks for this case, so we add this extra
    info to ExecutorOutput in memory.
    """
    for key, artifact_list in executor_output.artifacts.items():
        for artifact in artifact_list.artifacts:
            artifact.type.CopyFrom(
                executor_input.outputs.artifacts[key].artifacts[0].type)
    return executor_output


def get_outputs_for_task(
    executor_input: pipeline_spec_pb2.ExecutorInput,
    component_spec: pipeline_spec_pb2.ComponentSpec,
) -> Dict[str, Any]:
    """Gets outputs from a recently executed task, if available, using the
    ExecutorInput and ComponentSpec of the task."""
    executor_output = load_executor_output(
        executor_output_path=executor_input.outputs.output_file)
    return get_outputs_from_executor_output(
        executor_output=executor_output,
        executor_input=executor_input,
        component_spec=component_spec,
    )



# Copyright 2024 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Code for locally executing a DAG within a pipeline."""
import copy
from typing import Any, Dict, List, Tuple

from kfp.local import config
from kfp.local import graph_utils
from kfp.local import importer_handler
from kfp.local import io
from kfp.local import status
from kfp.pipeline_spec import pipeline_spec_pb2

Outputs = Dict[str, Any]


def run_dag(
    pipeline_resource_name: str,
    dag_component_spec: pipeline_spec_pb2.ComponentSpec,
    executors: Dict[str,
                    pipeline_spec_pb2.PipelineDeploymentConfig.ExecutorSpec],
    components: Dict[str, pipeline_spec_pb2.ComponentSpec],
    dag_arguments: Dict[str, Any],
    pipeline_root: str,
    runner: config.LocalRunnerType,
    unique_pipeline_id: str,
    fail_stack: List[str],
) -> Tuple[Outputs, status.Status]:
    """Runs a DAGSpec.

    Args:
        pipeline_resource_name: The root pipeline resource name.
        dag_component_spec: The ComponentSpec which defines the DAG to execute.
        executors: The ExecutorSpecs corresponding to the DAG.
        components: The ComponentSpecs corresponding to the DAG.
        dag_arguments: The arguments to the DAG's outer ComponentSpec.
        io_store: The IOStore instance corresponding to this DAG.
        pipeline_root: The local pipeline root.
        runner: The user-specified local runner.
        unique_pipeline_id: A unique identifier for the pipeline for placeholder resolution.
        fail_stack: Mutable stack of failures. If a primitive task in the DAG fails, the task name is appended. If a multi-task DAG fails, the DAG name is appended. If the pipeline executes successfully, fail_stack will be empty throughout the full local execution call stack.

    Returns:
        A two-tuple of (outputs, status). If status is FAILURE, outputs is an empty dictionary.
    """
    from kfp.local import task_dispatcher

    dag_arguments_with_defaults = join_user_inputs_and_defaults(
        dag_arguments=dag_arguments,
        dag_inputs_spec=dag_component_spec.input_definitions,
    )

    # prepare IOStore for DAG
    io_store = io.IOStore()
    for k, v in dag_arguments_with_defaults.items():
        io_store.put_parent_input(k, v)

    # execute tasks in order
    dag_spec = dag_component_spec.dag
    sorted_tasks = graph_utils.topological_sort_tasks(dag_spec.tasks)
    while sorted_tasks:
        task_name = sorted_tasks.pop()
        task_spec = dag_spec.tasks[task_name]
        # TODO: support control flow features
        validate_task_spec_not_loop_or_condition(task_spec=task_spec)
        component_name = task_spec.component_ref.name
        component_spec = components[component_name]
        implementation = component_spec.WhichOneof('implementation')
        if implementation == 'dag':
            # unlikely to exceed default max recursion depth of 1000
            outputs, task_status = run_dag(
                pipeline_resource_name=pipeline_resource_name,
                dag_component_spec=component_spec,
                components=components,
                executors=executors,
                dag_arguments=make_task_arguments(
                    task_spec.inputs,
                    io_store,
                ),
                pipeline_root=pipeline_root,
                runner=runner,
                unique_pipeline_id=unique_pipeline_id,
                fail_stack=fail_stack,
            )

        elif implementation == 'executor_label':
            executor_spec = executors[component_spec.executor_label]
            task_arguments = make_task_arguments(
                task_inputs_spec=dag_spec.tasks[task_name].inputs,
                io_store=io_store,
            )

            if executor_spec.WhichOneof('spec') == 'importer':
                outputs, task_status = importer_handler.run_importer(
                    pipeline_resource_name=pipeline_resource_name,
                    component_name=component_name,
                    component_spec=component_spec,
                    executor_spec=executor_spec,
                    arguments=task_arguments,
                    pipeline_root=pipeline_root,
                    unique_pipeline_id=unique_pipeline_id,
                )
            elif executor_spec.WhichOneof('spec') == 'container':
                outputs, task_status = task_dispatcher.run_single_task_implementation(
                    pipeline_resource_name=pipeline_resource_name,
                    component_name=component_name,
                    component_spec=component_spec,
                    executor_spec=executor_spec,
                    arguments=task_arguments,
                    pipeline_root=pipeline_root,
                    runner=runner,
                    # let the outer pipeline raise the error
                    raise_on_error=False,
                    # components may consume input artifacts when passed from upstream
                    # outputs or parent component inputs
                    block_input_artifact=False,
                    # provide the same unique job id for each task for
                    # consistent placeholder resolution
                    unique_pipeline_id=unique_pipeline_id,
                )
            else:
                raise ValueError(
                    "Got unknown spec in ExecutorSpec. Only 'dsl.component', 'dsl.container_component', and 'dsl.importer' are supported in local pipeline execution."
                )
        else:
            raise ValueError(
                f'Got unknown component implementation: {implementation}')

        if task_status == status.Status.FAILURE:
            fail_stack.append(task_name)
            return {}, status.Status.FAILURE

        # update IO store on success
        elif task_status == status.Status.SUCCESS:
            for key, output in outputs.items():
                io_store.put_task_output(
                    task_name,
                    key,
                    output,
                )
        else:
            raise ValueError(f'Got unknown task status: {task_status.name}')

    dag_outputs = get_dag_outputs(
        dag_outputs_spec=dag_component_spec.dag.outputs,
        io_store=io_store,
    )
    return dag_outputs, status.Status.SUCCESS


def join_user_inputs_and_defaults(
    dag_arguments: Dict[str, Any],
    dag_inputs_spec: pipeline_spec_pb2.ComponentInputsSpec,
) -> Dict[str, Any]:
    """Collects user-provided arguments and default arguments (when no user-
    provided argument) into a dictionary. Returns the dictionary.

    Args:
        dag_arguments: The user-provided arguments to the DAG.
        dag_inputs_spec: The ComponentInputSpec for the DAG.

    Returns:
        The complete DAG inputs, with defaults included where the user-provided argument is missing.
    """
    from kfp.local import executor_output_utils

    copied_dag_arguments = copy.deepcopy(dag_arguments)

    for input_name, input_spec in dag_inputs_spec.parameters.items():
        if input_name not in copied_dag_arguments:
            copied_dag_arguments[
                input_name] = executor_output_utils.pb2_value_to_python(
                    input_spec.default_value)
    return copied_dag_arguments


def make_task_arguments(
    task_inputs_spec: pipeline_spec_pb2.TaskInputsSpec,
    io_store: io.IOStore,
) -> Dict[str, Any]:
    """Obtains a dictionary of arguments required to execute the task
    corresponding to TaskInputsSpec.

    Args:
        task_inputs_spec: The TaskInputsSpec for the task.
        io_store: The IOStore of the current DAG. Used to obtain task arguments which come from upstream task outputs and parent component inputs.

    Returns:
        The arguments for the task.
    """
    from kfp.local import executor_output_utils

    task_arguments = {}
    # handle parameters
    for input_name, input_spec in task_inputs_spec.parameters.items():

        # handle constants
        if input_spec.HasField('runtime_value'):
            # runtime_value's value should always be constant for the v2 compiler
            if input_spec.runtime_value.WhichOneof('value') != 'constant':
                raise ValueError('Expected constant.')
            task_arguments[
                input_name] = executor_output_utils.pb2_value_to_python(
                    input_spec.runtime_value.constant)

        # handle upstream outputs
        elif input_spec.HasField('task_output_parameter'):
            task_arguments[input_name] = io_store.get_task_output(
                input_spec.task_output_parameter.producer_task,
                input_spec.task_output_parameter.output_parameter_key,
            )

        # handle parent pipeline input parameters
        elif input_spec.HasField('component_input_parameter'):
            task_arguments[input_name] = io_store.get_parent_input(
                input_spec.component_input_parameter)

        # TODO: support dsl.ExitHandler
        elif input_spec.HasField('task_final_status'):
            raise NotImplementedError(
                "'dsl.ExitHandler' is not yet support for local execution.")

        else:
            raise ValueError(f'Missing input for parameter {input_name}.')

    # handle artifacts
    for input_name, input_spec in task_inputs_spec.artifacts.items():
        if input_spec.HasField('task_output_artifact'):
            task_arguments[input_name] = io_store.get_task_output(
                input_spec.task_output_artifact.producer_task,
                input_spec.task_output_artifact.output_artifact_key,
            )
        elif input_spec.HasField('component_input_artifact'):
            task_arguments[input_name] = io_store.get_parent_input(
                input_spec.component_input_artifact)
        else:
            raise ValueError(f'Missing input for artifact {input_name}.')

    return task_arguments


def get_dag_output_parameters(
    dag_outputs_spec: pipeline_spec_pb2.DagOutputsSpec,
    io_store: io.IOStore,
) -> Dict[str, Any]:
    """Gets the DAG output parameter values from a DagOutputsSpec and the DAG's
    IOStore.

    Args:
        dag_outputs_spec: DagOutputsSpec corresponding to the DAG.
        io_store: IOStore corresponding to the DAG.

    Returns:
        The DAG output parameters.
    """
    outputs = {}
    for root_output_key, parameter_selector_spec in dag_outputs_spec.parameters.items(
    ):
        kind = parameter_selector_spec.WhichOneof('kind')
        if kind == 'value_from_parameter':
            value_from_parameter = parameter_selector_spec.value_from_parameter
            outputs[root_output_key] = io_store.get_task_output(
                value_from_parameter.producer_subtask,
                value_from_parameter.output_parameter_key,
            )
        elif kind == 'value_from_oneof':
            raise NotImplementedError(
                "'dsl.OneOf' is not yet supported in local execution.")
        else:
            raise ValueError(
                f"Got unknown 'parameter_selector_spec' kind: {kind}")
    return outputs


def get_dag_output_artifacts(
    dag_outputs_spec: pipeline_spec_pb2.DagOutputsSpec,
    io_store: io.IOStore,
) -> Dict[str, Any]:
    """Gets the DAG output artifact values from a DagOutputsSpec and the DAG's
    IOStore.

    Args:
        dag_outputs_spec: DagOutputsSpec corresponding to the DAG.
        io_store: IOStore corresponding to the DAG.

    Returns:
        The DAG output artifacts.
    """
    outputs = {}
    for root_output_key, artifact_selector_spec in dag_outputs_spec.artifacts.items(
    ):
        len_artifact_selectors = len(artifact_selector_spec.artifact_selectors)
        if len_artifact_selectors != 1:
            raise ValueError(
                f'Expected 1 artifact in ArtifactSelectorSpec. Got: {len_artifact_selectors}'
            )
        artifact_selector = artifact_selector_spec.artifact_selectors[0]
        outputs[root_output_key] = io_store.get_task_output(
            artifact_selector.producer_subtask,
            artifact_selector.output_artifact_key,
        )
    return outputs


def get_dag_outputs(
    dag_outputs_spec: pipeline_spec_pb2.DagOutputsSpec,
    io_store: io.IOStore,
) -> Dict[str, Any]:
    """Gets the DAG output values from a DagOutputsSpec and the DAG's IOStore.

    Args:
        dag_outputs_spec: DagOutputsSpec corresponding to the DAG.
        io_store: IOStore corresponding to the DAG.

    Returns:
        The DAG outputs.
    """
    output_params = get_dag_output_parameters(
        dag_outputs_spec=dag_outputs_spec,
        io_store=io_store,
    )
    output_artifacts = get_dag_output_artifacts(
        dag_outputs_spec=dag_outputs_spec,
        io_store=io_store,
    )
    return {**output_params, **output_artifacts}


def validate_task_spec_not_loop_or_condition(
        task_spec: pipeline_spec_pb2.PipelineTaskSpec) -> None:
    if task_spec.trigger_policy.condition:
        raise NotImplementedError(
            "'dsl.Condition' is not supported by local pipeline execution.")
    elif task_spec.WhichOneof('iterator'):
        raise NotImplementedError(
            "'dsl.ParallelFor' is not supported by local pipeline execution.")



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes and utilities for using and creating artifacts in components."""

import enum
import os
from typing import Dict, List, Optional, Type
import warnings

_GCS_LOCAL_MOUNT_PREFIX = '/gcs/'
_MINIO_LOCAL_MOUNT_PREFIX = '/minio/'
_S3_LOCAL_MOUNT_PREFIX = '/s3/'
_OCI_LOCAL_MOUNT_PREFIX = '/oci/'


class RemotePrefix(enum.Enum):
    GCS = 'gs://'
    MINIO = 'minio://'
    S3 = 's3://'
    OCI = 'oci://'


class Artifact:
    """Represents a generic machine learning artifact.

    This class and all artifact classes store the name, uri, and metadata for a machine learning artifact. Use this artifact type when an artifact does not fit into another more specific artifact type (e.g., ``Model``, ``Dataset``).

    Args:
        name: Name of the artifact.
        uri: The artifact's location on disk or cloud storage.
        metadata: Arbitrary key-value pairs about the artifact.

    Example:
      ::

        from kfp import dsl
        from kfp.dsl import Output, Artifact, Input


        @dsl.component
        def create_artifact(
            data: str,
            output_artifact: Output[Artifact],
        ):
            with open(output_artifact.path, 'w') as f:
                f.write(data)


        @dsl.component
        def use_artifact(input_artifact: Input[Artifact]):
            with open(input_artifact.path) as input_file:
                artifact_contents = input_file.read()
                print(artifact_contents)


        @dsl.pipeline(name='my-pipeline', pipeline_root='gs://my/storage')
        def my_pipeline():
            create_task = create_artifact(data='my data')
            use_artifact(input_artifact=create_task.outputs['output_artifact'])

    Note: Other artifacts are used similarly to the usage of ``Artifact`` in the example above (within ``Input[]`` and ``Output[]``).
    """
    schema_title = 'system.Artifact'
    schema_version = '0.0.1'

    def __init__(self,
                 name: Optional[str] = None,
                 uri: Optional[str] = None,
                 metadata: Optional[Dict] = None) -> None:
        """Initializes the Artifact with the given name, URI and metadata."""
        self.uri = uri or ''
        self.name = name or ''
        self.metadata = metadata or {}

    @property
    def path(self) -> str:
        return self._get_path()

    @path.setter
    def path(self, path: str) -> None:
        self._set_path(path)

    def _get_path(self) -> Optional[str]:
        if self.uri.startswith(RemotePrefix.GCS.value):
            return _GCS_LOCAL_MOUNT_PREFIX + self.uri[len(RemotePrefix.GCS.value
                                                         ):]
        if self.uri.startswith(RemotePrefix.MINIO.value):
            return _MINIO_LOCAL_MOUNT_PREFIX + self.uri[len(RemotePrefix.MINIO
                                                            .value):]
        if self.uri.startswith(RemotePrefix.S3.value):
            return _S3_LOCAL_MOUNT_PREFIX + self.uri[len(RemotePrefix.S3.value
                                                        ):]
        if self.uri.startswith(RemotePrefix.OCI.value):
            escaped_uri = self.uri[len(RemotePrefix.OCI.value):].replace(
                '/', '_')
            return _OCI_LOCAL_MOUNT_PREFIX + escaped_uri
        # uri == path for local execution
        return self.uri

    def _set_path(self, path: str) -> None:
        self.uri = convert_local_path_to_remote_path(path)


def convert_local_path_to_remote_path(path: str) -> str:
    if path.startswith(_GCS_LOCAL_MOUNT_PREFIX):
        return RemotePrefix.GCS.value + path[len(_GCS_LOCAL_MOUNT_PREFIX):]
    elif path.startswith(_MINIO_LOCAL_MOUNT_PREFIX):
        return RemotePrefix.MINIO.value + path[len(_MINIO_LOCAL_MOUNT_PREFIX):]
    elif path.startswith(_S3_LOCAL_MOUNT_PREFIX):
        return RemotePrefix.S3.value + path[len(_S3_LOCAL_MOUNT_PREFIX):]
    elif path.startswith(_OCI_LOCAL_MOUNT_PREFIX):
        remote_path = path[len(_OCI_LOCAL_MOUNT_PREFIX):].replace('_', '/')
        if remote_path.endswith("/models"):
            remote_path = remote_path[:-len("/models")]

        return RemotePrefix.OCI.value + remote_path

    return path


class Model(Artifact):
    """An artifact representing a machine learning model.

    Args:
        name: Name of the model.
        uri: The model's location on disk or cloud storage.
        metadata: Arbitrary key-value pairs about the model.
    """
    schema_title = 'system.Model'

    @property
    def framework(self) -> str:
        return self._get_framework()

    def _get_framework(self) -> str:
        return self.metadata.get('framework', '')

    @property
    def path(self) -> str:
        if self.uri.startswith("oci://"):
            # Modelcar container images are expected to have the model files stored in /models
            # https://github.com/kserve/kserve/blob/v0.14.1/pkg/webhook/admission/pod/storage_initializer_injector.go#L732
            return self._get_path() + "/models"

        return self._get_path()

    @framework.setter
    def framework(self, framework: str) -> None:
        self._set_framework(framework)

    def _set_framework(self, framework: str) -> None:
        self.metadata['framework'] = framework


class Dataset(Artifact):
    """An artifact representing a machine learning dataset.

    Args:
        name: Name of the dataset.
        uri: The dataset's location on disk or cloud storage.
        metadata: Arbitrary key-value pairs about the dataset.
    """
    schema_title = 'system.Dataset'


class Metrics(Artifact):
    """An artifact for storing key-value scalar metrics.

    Args:
        name: Name of the metrics artifact.
        uri: The metrics artifact's location on disk or cloud storage.
        metadata: Key-value scalar metrics.
    """
    schema_title = 'system.Metrics'

    def log_metric(self, metric: str, value: float) -> None:
        """Sets a custom scalar metric in the artifact's metadata.

        Args:
          metric: The metric key.
          value: The metric value.
        """
        self.metadata[metric] = value


class ClassificationMetrics(Artifact):
    """An artifact for storing classification metrics.

    Args:
        name: Name of the metrics artifact.
        uri: The metrics artifact's location on disk or cloud storage.
        metadata: The key-value scalar metrics.
    """
    schema_title = 'system.ClassificationMetrics'

    def log_roc_data_point(self, fpr: float, tpr: float,
                           threshold: float) -> None:
        """Logs a single data point in the ROC curve to metadata.

        Args:
          fpr: False positive rate value of the data point.
          tpr: True positive rate value of the data point.
          threshold: Threshold value for the data point.
        """

        roc_reading = {
            'confidenceThreshold': threshold,
            'recall': tpr,
            'falsePositiveRate': fpr
        }
        if 'confidenceMetrics' not in self.metadata.keys():
            self.metadata['confidenceMetrics'] = []

        self.metadata['confidenceMetrics'].append(roc_reading)

    def log_roc_curve(self, fpr: List[float], tpr: List[float],
                      threshold: List[float]) -> None:
        """Logs an ROC curve to metadata.

        Args:
          fpr: List of false positive rate values.
          tpr: List of true positive rate values.
          threshold: List of threshold values.

        Raises:
          ValueError: If the lists ``fpr``, ``tpr`` and ``threshold`` are not the same length.
        """
        if len(fpr) != len(tpr) or len(fpr) != len(threshold) or len(
                tpr) != len(threshold):
            raise ValueError(
                f'Length of fpr, tpr and threshold must be the same. Got lengths {len(fpr)}, {len(tpr)} and {len(threshold)} respectively.'
            )

        for i in range(len(fpr)):
            self.log_roc_data_point(
                fpr=fpr[i], tpr=tpr[i], threshold=threshold[i])

    def set_confusion_matrix_categories(self, categories: List[str]) -> None:
        """Stores confusion matrix categories to metadata.

        Args:
          categories: List of strings specifying the categories.
        """

        self._categories = []
        annotation_specs = []
        for category in categories:
            annotation_spec = {'displayName': category}
            self._categories.append(category)
            annotation_specs.append(annotation_spec)

        self._matrix = []
        for row in range(len(self._categories)):
            self._matrix.append({'row': [0] * len(self._categories)})

        self._confusion_matrix = {
            'annotationSpecs': annotation_specs,
            'rows': self._matrix
        }

        self.metadata['confusionMatrix'] = self._confusion_matrix

    def log_confusion_matrix_row(self, row_category: str,
                                 row: List[float]) -> None:
        """Logs a confusion matrix row to metadata.

        Args:
          row_category: Category to which the row belongs.
          row: List of integers specifying the values for the row.

        Raises:
          ValueError: If ``row_category`` is not in the list of categories
            set in ``set_categories`` call.
        """
        if row_category not in self._categories:
            raise ValueError(
                f'Invalid category: {row_category} passed. Expected one of: {self._categories}'
            )

        if len(row) != len(self._categories):
            raise ValueError(
                f'Invalid row. Expected size: {len(self._categories)} got: {len(row)}'
            )

        self._matrix[self._categories.index(row_category)] = {'row': row}
        self.metadata['confusionMatrix'] = self._confusion_matrix

    def log_confusion_matrix_cell(self, row_category: str, col_category: str,
                                  value: int) -> None:
        """Logs a cell in the confusion matrix to metadata.

        Args:
          row_category: String representing the name of the row category.
          col_category: String representing the name of the column category.
          value: Value of the cell.

        Raises:
          ValueError: If ``row_category`` or ``col_category`` is not in the list of
           categories set in ``set_categories``.
        """
        if row_category not in self._categories:
            raise ValueError(
                f'Invalid category: {row_category} passed. Expected one of: {self._categories}'
            )

        if col_category not in self._categories:
            raise ValueError(
                f'Invalid category: {row_category} passed. Expected one of: {self._categories}'
            )

        self._matrix[self._categories.index(row_category)]['row'][
            self._categories.index(col_category)] = value
        self.metadata['confusionMatrix'] = self._confusion_matrix

    def log_confusion_matrix(self, categories: List[str],
                             matrix: List[List[int]]) -> None:
        """Logs a confusion matrix to metadata.

        Args:
          categories: List of the category names.
          matrix: Complete confusion matrix.

        Raises:
          ValueError: If the length of ``categories`` does not match number of rows or columns of ``matrix``.
        """
        self.set_confusion_matrix_categories(categories)

        if len(matrix) != len(categories):
            raise ValueError(
                f'Invalid matrix: {matrix} passed for categories: {categories}')

        for index in range(len(categories)):
            if len(matrix[index]) != len(categories):
                raise ValueError(
                    f'Invalid matrix: {matrix} passed for categories: {categories}'
                )

            self.log_confusion_matrix_row(categories[index], matrix[index])

        self.metadata['confusionMatrix'] = self._confusion_matrix


class SlicedClassificationMetrics(Artifact):
    """An artifact for storing sliced classification metrics.

    Similar to ``ClassificationMetrics``, tasks using this class are
    expected to use log methods of the class to log metrics with the
    difference being each log method takes a slice to associate the
    ``ClassificationMetrics``.

    Args:
        name: Name of the metrics artifact.
        uri: The metrics artifact's location on disk or cloud storage.
        metadata: Arbitrary key-value pairs about the metrics artifact.
    """

    schema_title = 'system.SlicedClassificationMetrics'

    def _upsert_classification_metrics_for_slice(self, slice: str) -> None:
        """Upserts the classification metrics instance for a slice."""
        if slice not in self._sliced_metrics:
            self._sliced_metrics[slice] = ClassificationMetrics()

    def _update_metadata(self, slice: str) -> None:
        """Updates metadata to adhere to the metrics schema."""
        self.metadata = {'evaluationSlices': []}
        for slice in self._sliced_metrics.keys():
            slice_metrics = {
                'slice':
                    slice,
                'sliceClassificationMetrics':
                    self._sliced_metrics[slice].metadata
            }
            self.metadata['evaluationSlices'].append(slice_metrics)

    def log_roc_reading(self, slice: str, threshold: float, tpr: float,
                        fpr: float) -> None:
        """Logs a single data point in the ROC curve of a slice to metadata.

        Args:
          slice: String representing slice label.
          threshold: Thresold value for the data point.
          tpr: True positive rate value of the data point.
          fpr: False positive rate value of the data point.
        """

        self._upsert_classification_metrics_for_slice(slice)
        self._sliced_metrics[slice].log_roc_reading(threshold, tpr, fpr)
        self._update_metadata(slice)

    def load_roc_readings(self, slice: str,
                          readings: List[List[float]]) -> None:
        """Bulk loads ROC curve readings for a slice.

        Args:
          slice: String representing slice label.
          readings: A 2-dimensional list providing ROC curve data points. The expected order of the data points is: threshold, true positive rate, false positive rate.
        """
        self._upsert_classification_metrics_for_slice(slice)
        self._sliced_metrics[slice].load_roc_readings(readings)
        self._update_metadata(slice)

    def set_confusion_matrix_categories(self, slice: str,
                                        categories: List[str]) -> None:
        """Logs confusion matrix categories for a slice to metadata.

        Categories are stored in the internal ``metrics_utils.ConfusionMatrix``
        instance of the slice.

        Args:
          slice: String representing slice label.
          categories: List of strings specifying the categories.
        """
        self._upsert_classification_metrics_for_slice(slice)
        self._sliced_metrics[slice].set_confusion_matrix_categories(categories)
        self._update_metadata(slice)

    def log_confusion_matrix_row(self, slice: str, row_category: str,
                                 row: List[int]) -> None:
        """Logs a confusion matrix row for a slice to metadata.

        Row is updated on the internal ``metrics_utils.ConfusionMatrix``
        instance of the slice.

        Args:
          slice: String representing slice label.
          row_category: Category to which the row belongs.
          row: List of integers specifying the values for the row.
        """
        self._upsert_classification_metrics_for_slice(slice)
        self._sliced_metrics[slice].log_confusion_matrix_row(row_category, row)
        self._update_metadata(slice)

    def log_confusion_matrix_cell(self, slice: str, row_category: str,
                                  col_category: str, value: int) -> None:
        """Logs a confusion matrix cell for a slice to metadata.

        Cell is updated on the internal ``metrics_utils.ConfusionMatrix``
        instance of the slice.

        Args:
          slice: String representing slice label.
          row_category: String representing the name of the row category.
          col_category: String representing the name of the column category.
          value: Value of the cell.
        """
        self._upsert_classification_metrics_for_slice(slice)
        self._sliced_metrics[slice].log_confusion_matrix_cell(
            row_category, col_category, value)
        self._update_metadata(slice)

    def load_confusion_matrix(self, slice: str, categories: List[str],
                              matrix: List[List[int]]) -> None:
        """Bulk loads the whole confusion matrix for a slice.

        Args:
          slice: String representing slice label.
          categories: List of the category names.
          matrix: Complete confusion matrix.
        """
        self._upsert_classification_metrics_for_slice(slice)
        self._sliced_metrics[slice].log_confusion_matrix_cell(
            categories, matrix)
        self._update_metadata(slice)


class HTML(Artifact):
    """An artifact representing an HTML file.

    Args:
        name: Name of the HTML file.
        uri: The HTML file's location on disk or cloud storage.
        metadata: Arbitrary key-value pairs about the HTML file.
    """
    schema_title = 'system.HTML'


class Markdown(Artifact):
    """An artifact representing a markdown file.

    Args:
        name: Name of the markdown file.
        uri: The markdown file's location on disk or cloud storage.
        metadata: Arbitrary key-value pairs about the markdown file.
    """
    schema_title = 'system.Markdown'


_SCHEMA_TITLE_TO_TYPE: Dict[str, Type[Artifact]] = {
    x.schema_title: x for x in [
        Artifact,
        Model,
        Dataset,
        Metrics,
        ClassificationMetrics,
        SlicedClassificationMetrics,
        HTML,
        Markdown,
    ]
}

CONTAINER_TASK_ROOT: Optional[str] = None


# suffix default of 'Output' should be the same key as the default key for a
# single output component, but use value not variable for reference docs
def get_uri(suffix: str = 'Output') -> str:
    """Gets the task root URI, a unique object storage URI associated with the
    current task. This function may only be called at task runtime.

    Returns an empty string if the task root cannot be inferred from the runtime environment.

    Args:
        suffix: A suffix to append to the URI. This is a helpful for creating unique subdirectories when the component has multiple outputs.

    Returns:
        The URI or empty string.
    """
    if CONTAINER_TASK_ROOT is None:
        raise RuntimeError(
            f"'dsl.{get_uri.__name__}' can only be called at task runtime. The task root is unknown in the current environment."
        )
    UNSUPPORTED_KFP_PATH = '/tmp/kfp_outputs'
    if CONTAINER_TASK_ROOT == UNSUPPORTED_KFP_PATH:
        warnings.warn(
            f'dsl.{get_uri.__name__} is not yet supported by the KFP backend. Please specify a URI explicitly.',
            RuntimeWarning,
            stacklevel=2,
        )
        # return empty string, not None, to conform with logic in artifact
        # constructor which immediately converts uri=None to uri=''
        # this way the .path property can worry about handling fewer input types
        return ''
    remote_task_root = convert_local_path_to_remote_path(CONTAINER_TASK_ROOT)
    return os.path.join(remote_task_root, suffix)



# Copyright 2023 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implementation of the subprocess runner."""

import contextlib
import os
import subprocess
import sys
import tempfile
from typing import List
import venv
import warnings

from kfp.dsl import component_factory
from kfp.local import config
from kfp.local import status
from kfp.local import task_handler_interface


class SubprocessTaskHandler(task_handler_interface.ITaskHandler):
    """The task handler corresponding to kfp.local.SubprocessRunner."""

    def __init__(
        self,
        image: str,
        full_command: List[str],
        pipeline_root: str,
        runner: config.SubprocessRunner,
    ) -> None:
        self.validate_image(image)
        self.validate_not_container_component(full_command)
        self.validate_not_containerized_python_component(full_command)

        self.image = image
        self.full_command = full_command
        self.pipeline_root = pipeline_root
        self.runner = runner

    def run(self) -> status.Status:
        """Runs the local subprocess and returns the status.

        Returns:
            Status.
        """
        with environment(use_venv=self.runner.use_venv) as py_executable:
            full_command = replace_python_executable(
                self.full_command,
                py_executable,
            )
            return_code = run_local_subprocess(full_command=full_command)
            return status.Status.SUCCESS if return_code == 0 else status.Status.FAILURE

    def validate_image(self, image: str) -> None:
        if 'python' not in image:
            warnings.warn(
                f"You may be attemping to run a task that uses custom or non-Python base image '{image}' in a Python environment. This may result in incorrect dependencies and/or incorrect behavior. Consider using the 'DockerRunner' to run this task in a container.",
                RuntimeWarning,
            )

    def validate_not_container_component(
        self,
        full_command: List[str],
    ) -> None:
        if not any(component_factory.EXECUTOR_MODULE in part
                   for part in full_command):
            raise RuntimeError(
                f'The {config.SubprocessRunner.__name__} only supports running Lightweight Python Components. You are attempting to run a Container Component.'
            )

    def validate_not_containerized_python_component(
        self,
        full_command: List[str],
    ) -> None:
        if full_command[:len(
                component_factory.CONTAINERIZED_PYTHON_COMPONENT_COMMAND
        )] == component_factory.CONTAINERIZED_PYTHON_COMPONENT_COMMAND:
            raise RuntimeError(
                f'The {config.SubprocessRunner.__name__} only supports running Lightweight Python Components. You are attempting to run a Containerized Python Component.'
            )


def run_local_subprocess(full_command: List[str]) -> int:
    with subprocess.Popen(
            full_command,
            stdout=subprocess.PIPE,
            # No change to behavior in terminal for user,
            # but inner process logs redirected to stdout. This separates from
            # the outer process logs which, per logging module default, go to
            # stderr.
            stderr=subprocess.STDOUT,
            text=True,
            # buffer line-by-line
            bufsize=1,
    ) as process:
        if process.stdout:
            for line in iter(process.stdout.readline, ''):
                print(line, end='')

            # help with visual separation to show termination of subprocess logs
            print('\n')

        return process.wait()


def replace_python_executable(full_command: List[str],
                              new_executable: str) -> List[str]:
    """Replaces the 'python3' string in each element of the full_command with
    the new_executable.

    Args:
        full_command: Commands and args.
        new_executable: The Python executable to use for local execution.

    Returns:
        The updated commands and args.
    """
    return [el.replace('python3', f'{new_executable}') for el in full_command]


@contextlib.contextmanager
def environment(use_venv: bool) -> str:
    """Context manager that handles the environment used for the subprocess.

    Args:
        use_venv: Whether to use the virtual environment instead of current environment.

    Returns:
        The Python executable path to use.
    """
    if use_venv:
        with tempfile.TemporaryDirectory() as tempdir:
            # Create the virtual environment inside the temporary directory
            venv.create(tempdir, with_pip=True)

            yield os.path.join(tempdir, 'bin', 'python')
    else:
        yield sys.executable



# Copyright 2024 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Pipeline-level config options."""

from typing import Any, Dict, Optional


class KubernetesWorkspaceConfig:
    """Configuration for Kubernetes-specific workspace settings.

    Use this to override the default PersistentVolumeClaim (PVC) configuration
    used when running pipelines on a Kubernetes cluster.

    Attributes:
        pvcSpecPatch: A dictionary of fields to patch onto the default PVC spec
                      (e.g., 'storageClassName', 'accessModes').
    """

    def __init__(self, pvcSpecPatch: Optional[Dict[str, Any]] = None):
        self.pvcSpecPatch = pvcSpecPatch or {}

    def set_pvcSpecPatch(self, patch: Dict[str, Any]):
        self.pvcSpecPatch = patch


class WorkspaceConfig:
    """Configuration for a shared workspace that persists during the pipeline
    run.

    Attributes:
        size (str): The size of the workspace (e.g., '250Gi'). This is a required field.
        See https://kubernetes.io/docs/reference/kubernetes-api/common-definitions/quantity/ for valid quantity formats.
        kubernetes: (Optional) Kubernetes-specific configuration for the underlying PVC.
    """

    def __init__(self,
                 size: str,
                 kubernetes: Optional[KubernetesWorkspaceConfig] = None):
        if not size or not size.strip():
            raise ValueError('Workspace size is required and cannot be empty')
        self.size = size.strip()
        self.kubernetes = kubernetes or KubernetesWorkspaceConfig()

    def get_workspace(self) -> dict:
        workspace = {'size': self.size}
        if self.kubernetes:
            workspace['kubernetes'] = {
                'pvcSpecPatch': self.kubernetes.pvcSpecPatch
            }
        return workspace

    def set_size(self, size: str):
        if not size or not size.strip():
            raise ValueError('Workspace size is required and cannot be empty')
        self.size = size.strip()

    def set_kubernetes_config(self,
                              kubernetes_config: KubernetesWorkspaceConfig):
        self.kubernetes = kubernetes_config


class PipelineConfig:
    """PipelineConfig contains pipeline-level config options."""

    def __init__(self, workspace: Optional[WorkspaceConfig] = None):
        self.workspace = workspace



# Copyright 2023 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for executor_input_utils.py."""

import unittest

from google.protobuf import json_format
from kfp import dsl
from kfp.local import executor_input_utils
from kfp.local import testing_utilities
from kfp.pipeline_spec import pipeline_spec_pb2


class GetLocalPipelineResourceName(testing_utilities.MockedDatetimeTestCase):

    def test(self):
        actual = executor_input_utils.get_local_pipeline_resource_name(
            'my-pipeline')
        expected = 'my-pipeline-2023-10-10-13-32-59-420710'
        self.assertEqual(actual, expected)


class GetLocalTaskResourceName(unittest.TestCase):

    def test(self):
        actual = executor_input_utils.get_local_task_resource_name(
            'comp-my-comp')
        expected = 'my-comp'
        self.assertEqual(actual, expected)


class TestConstructLocalTaskRoot(testing_utilities.MockedDatetimeTestCase):

    def test(self):

        task_root = executor_input_utils.construct_local_task_root(
            pipeline_root='/foo/bar',
            pipeline_resource_name='my-pipeline-2023-10-10-13-32-59-420710',
            task_resource_name='my-comp',
        )
        self.assertEqual(
            task_root,
            '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/my-comp',
        )


class TestConstructExecutorInput(unittest.TestCase):

    def test_no_inputs(self):
        component_spec = pipeline_spec_pb2.ComponentSpec()
        json_format.ParseDict(
            {
                'outputDefinitions': {
                    'parameters': {
                        'Output': {
                            'parameterType': 'STRING'
                        }
                    }
                },
                'executorLabel': 'exec-comp'
            }, component_spec)
        arguments = {}
        task_root = '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp'

        actual = executor_input_utils.construct_executor_input(
            component_spec=component_spec,
            arguments=arguments,
            task_root=task_root,
            block_input_artifact=True,
        )
        expected = pipeline_spec_pb2.ExecutorInput()
        json_format.ParseDict(
            {
                'inputs': {},
                'outputs': {
                    'parameters': {
                        'Output': {
                            'outputFile':
                                '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp/Output'
                        }
                    },
                    'outputFile':
                        '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp/executor_output.json'
                }
            }, expected)
        self.assertEqual(actual, expected)

    def test_various_io_types(self):
        component_spec = pipeline_spec_pb2.ComponentSpec()
        json_format.ParseDict(
            {
                'inputDefinitions': {
                    'parameters': {
                        'boolean': {
                            'parameterType': 'BOOLEAN'
                        }
                    }
                },
                'outputDefinitions': {
                    'artifacts': {
                        'out_a': {
                            'artifactType': {
                                'schemaTitle': 'system.Dataset',
                                'schemaVersion': '0.0.1'
                            }
                        }
                    },
                    'parameters': {
                        'Output': {
                            'parameterType': 'NUMBER_INTEGER'
                        }
                    }
                },
                'executorLabel': 'exec-comp'
            }, component_spec)
        arguments = {'boolean': False}
        task_root = '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp'

        actual = executor_input_utils.construct_executor_input(
            component_spec=component_spec,
            arguments=arguments,
            task_root=task_root,
            block_input_artifact=True,
        )
        expected = pipeline_spec_pb2.ExecutorInput()
        json_format.ParseDict(
            {
                'inputs': {
                    'parameterValues': {
                        'boolean': False
                    }
                },
                'outputs': {
                    'parameters': {
                        'Output': {
                            'outputFile':
                                '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp/Output'
                        }
                    },
                    'artifacts': {
                        'out_a': {
                            'artifacts': [{
                                'name':
                                    'out_a',
                                'type': {
                                    'schemaTitle': 'system.Dataset',
                                    'schemaVersion': '0.0.1'
                                },
                                'uri':
                                    '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp/out_a',
                                'metadata': {}
                            }]
                        }
                    },
                    'outputFile':
                        '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp/executor_output.json'
                }
            }, expected)
        self.assertEqual(actual, expected)

    def test_block_input_artifact(self):
        component_spec = pipeline_spec_pb2.ComponentSpec()
        json_format.ParseDict(
            {
                'inputDefinitions': {
                    'artifacts': {
                        'in_artifact': {
                            'artifactType': {
                                'schemaTitle': 'system.Artifact',
                                'schemaVersion': '0.0.1'
                            }
                        }
                    }
                },
                'executorLabel': 'exec-comp'
            }, component_spec)
        arguments = {}
        task_root = '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp'
        with self.assertRaisesRegex(
                ValueError,
                'Input artifacts are not yet supported for local execution.'):
            executor_input_utils.construct_executor_input(
                component_spec=component_spec,
                arguments=arguments,
                task_root=task_root,
                block_input_artifact=True,
            )

    def test_allow_input_artifact(self):
        component_spec = pipeline_spec_pb2.ComponentSpec()
        json_format.ParseDict(
            {
                'inputDefinitions': {
                    'artifacts': {
                        'in_artifact': {
                            'artifactType': {
                                'schemaTitle': 'system.Artifact',
                                'schemaVersion': '0.0.1'
                            }
                        }
                    }
                },
                'executorLabel': 'exec-comp'
            }, component_spec)
        task_root = '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp'
        arguments = {
            'in_artifact':
                dsl.Artifact(
                    name='artifact',
                    uri='/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/prev-comp/artifact',
                    metadata={'foo': 'bar'})
        }
        actual = executor_input_utils.construct_executor_input(
            component_spec=component_spec,
            arguments=arguments,
            task_root=task_root,
            # this param says input artifacts should be permitted
            block_input_artifact=False,
        )
        expected = pipeline_spec_pb2.ExecutorInput()
        json_format.ParseDict(
            {
                'inputs': {
                    'artifacts': {
                        'in_artifact': {
                            'artifacts': [{
                                'name':
                                    'artifact',
                                'type': {
                                    'schemaTitle': 'system.Artifact',
                                    'schemaVersion': '0.0.1'
                                },
                                'uri':
                                    '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/prev-comp/artifact',
                                'metadata': {
                                    'foo': 'bar'
                                }
                            }]
                        }
                    }
                },
                'outputs': {
                    'outputFile':
                        '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp/executor_output.json'
                }
            }, expected)
        self.assertEqual(actual, expected)

    def test_fstring_case(self):
        component_spec = pipeline_spec_pb2.ComponentSpec()
        json_format.ParseDict(
            {
                'inputDefinitions': {
                    'parameters': {
                        'string': {
                            'parameterType': 'STRING'
                        }
                    }
                },
                'outputDefinitions': {
                    'parameters': {
                        'Output': {
                            'parameterType': 'STRING'
                        }
                    }
                },
                'executorLabel': 'exec-identity'
            }, component_spec)
        expected_executor_input = pipeline_spec_pb2.ExecutorInput()
        json_format.ParseDict(
            {
                'inputs': {
                    'parameterValues': {
                        'pipelinechannel--string':
                            'baz',
                        'string':
                            "bar-{{$.inputs.parameters['pipelinechannel--string']}}"
                    }
                },
                'outputs': {
                    'parameters': {
                        'Output': {
                            'outputFile':
                                '/foo/bar/local_outputs/my-pipeline-2024-01-26-11-10-57XX-530768/identity/Output'
                        }
                    },
                    'outputFile':
                        '/foo/bar/local_outputs/my-pipeline-2024-01-26-11-10-57XX-530768/identity/executor_output.json'
                }
            }, expected_executor_input)
        actual_executor_input = executor_input_utils.construct_executor_input(
            component_spec=component_spec,
            arguments={
                'pipelinechannel--string':
                    'baz',
                # covers the case of an f-string, where the value of
                # string includes an interpolation of
                # pipelinechannel--string
                'string':
                    "bar-{{$.inputs.parameters['pipelinechannel--string']}}"
            },
            task_root='/foo/bar/local_outputs/my-pipeline-2024-01-26-11-10-57XX-530768/identity',
            block_input_artifact=True,
        )
        self.assertEqual(
            expected_executor_input,
            actual_executor_input,
        )


class TestExecutorInputToDict(unittest.TestCase):

    def test_with_ints_and_floats(self):
        component_spec = pipeline_spec_pb2.ComponentSpec()
        json_format.ParseDict(
            {
                'inputDefinitions': {
                    'parameters': {
                        'x': {
                            'parameterType': 'NUMBER_INTEGER'
                        },
                        'y': {
                            'parameterType': 'NUMBER_DOUBLE'
                        }
                    }
                },
                'outputDefinitions': {
                    'parameters': {
                        'Output': {
                            'parameterType': 'STRING'
                        }
                    }
                },
                'executorLabel': 'exec-comp'
            }, component_spec)

        executor_input = pipeline_spec_pb2.ExecutorInput()
        json_format.ParseDict(
            {
                'inputs': {
                    'parameterValues': {
                        'x': 1.0,
                        'y': 2.0
                    }
                },
                'outputs': {
                    'parameters': {
                        'Output': {
                            'outputFile':
                                '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp/Output'
                        }
                    },
                    'outputFile':
                        '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp/executor_output.json'
                }
            }, executor_input)

        executor_input_dict = executor_input_utils.executor_input_to_dict(
            executor_input=executor_input,
            component_spec=component_spec,
        )
        expected = {
            'inputs': {
                'parameterValues': {
                    'x': 1,
                    'y': 2.0
                }
            },
            'outputs': {
                'parameters': {
                    'Output': {
                        'outputFile':
                            '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp/Output'
                    }
                },
                'outputFile':
                    '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp/executor_output.json'
            }
        }
        # assert types since 1.0 == 1
        self.assertIsInstance(
            executor_input_dict['inputs']['parameterValues']['x'], int)
        self.assertIsInstance(
            executor_input_dict['inputs']['parameterValues']['y'], float)
        self.assertEqual(executor_input_dict, expected)


if __name__ == '__main__':
    unittest.main()



# Copyright 2018-2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# `kfp` is a namespace package.
# https://packaging.python.org/guides/packaging-namespace-packages/#pkgutil-style-namespace-packages
__path__ = __import__('pkgutil').extend_path(__path__, __name__)

try:
    from .version import __version__
except ImportError:
    __version__ = 'dev'

import sys
import warnings

if sys.version_info < (3, 9):
    warnings.warn(
        ('KFP will drop support for Python 3.9 on Oct 1, 2025. To use new versions of the KFP SDK after that date, you will need to upgrade to Python >= 3.10. See https://devguide.python.org/versions/ for more details.'
        ),
        FutureWarning,
        stacklevel=2,
    )

TYPE_CHECK = True

import os

# compile-time only dependencies
if os.environ.get('_KFP_RUNTIME', 'false') != 'true':
    # make `from kfp import components` and `from kfp import dsl` valid;
    # related to namespace packaging issue
    from kfp import components  # noqa: keep unused import
    from kfp import dsl  # noqa: keep unused import
    from kfp.client import Client  # noqa: keep unused import



# Copyright 2024 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Graph algorithms which are useful for working with PipelineSpec."""

from typing import Dict, List, Set

from kfp.pipeline_spec import pipeline_spec_pb2


def topological_sort_tasks(
        tasks: Dict[str, pipeline_spec_pb2.PipelineTaskSpec]) -> List[str]:
    """Given a dictionary of task name to PipelineTaskSpec, obtains a
    topologically sorted stack of task names.

    Args:
        tasks: The tasks in the pipeline.

    Returns:
        A totally ordered stack of tasks. Tasks should be executed in the order they are popped off the right side of the stack.
    """
    dependency_map = build_dependency_map(tasks)
    return topological_sort(dependency_map)


def build_dependency_map(
    tasks: Dict[str,
                pipeline_spec_pb2.PipelineTaskSpec]) -> Dict[str, List[str]]:
    """Builds a dictionary of task name to all upstream task names
    (dependencies). This is a data structure simplification step, which allows
    for a general topological_sort sort implementation.

    Args:
        tasks: The tasks in the pipeline.

    Returns:
        An dictionary of task name to all upstream tasks. The key task depends on all value tasks being executed first.
    """
    return {
        task_name: task_details.dependent_tasks
        for task_name, task_details in tasks.items()
    }


def topological_sort(dependency_map: Dict[str, List[str]]) -> List[str]:
    """Topologically sorts a dictionary of task names to upstream tasks.

    Args:
        dependency_map: A dictionary of tasks name to a list of upstream tasks. The key task depends on all value tasks being executed first.

    Returns:
        A totally ordered stack of tasks. Tasks should be executed in the order they are popped off the right side of the stack.
    """

    def dfs(node: str) -> None:
        visited.add(node)
        for neighbor in dependency_map[node]:
            if neighbor not in visited:
                dfs(neighbor)
        result.append(node)

    # sort lists to force deterministic result
    dependency_map = {k: sorted(v) for k, v in dependency_map.items()}
    visited: Set[str] = set()
    result = []
    for node in dependency_map:
        if node not in visited:
            dfs(node)
    return result[::-1]



# Copyright 2024 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test for utils.py."""
import unittest

from google.protobuf import json_format
from google.protobuf import struct_pb2
from kfp.local import utils
from kfp.pipeline_spec import pipeline_spec_pb2


class TestDictToExecutorSpec(unittest.TestCase):

    def test_simple(self):
        input_struct = struct_pb2.Struct()
        input_dict = {
            'container': {
                'image': 'alpine',
                'command': ['echo'],
                'args': ['foo'],
            }
        }
        json_format.ParseDict(input_dict, input_struct)
        expected = pipeline_spec_pb2.PipelineDeploymentConfig.ExecutorSpec(
            container=pipeline_spec_pb2.PipelineDeploymentConfig
            .PipelineContainerSpec(
                image='alpine', command=['echo'], args=['foo']))

        actual = utils.struct_to_executor_spec(input_struct)
        self.assertEqual(actual, expected)


if __name__ == '__main__':
    unittest.main()



# Copyright 2024 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for importer_handler_test.py."""
import unittest

from google.protobuf import json_format
from kfp import dsl
from kfp.local import importer_handler
from kfp.local import status
from kfp.local import testing_utilities
from kfp.pipeline_spec import pipeline_spec_pb2


class TestRunImporter(testing_utilities.LocalRunnerEnvironmentTestCase):

    def test_uri_from_upstream(self):
        component_spec_dict = {
            'inputDefinitions': {
                'parameters': {
                    'metadata': {
                        'parameterType': 'STRING'
                    },
                    'uri': {
                        'parameterType': 'STRING'
                    }
                }
            },
            'outputDefinitions': {
                'artifacts': {
                    'artifact': {
                        'artifactType': {
                            'schemaTitle': 'system.Dataset',
                            'schemaVersion': '0.0.1'
                        }
                    }
                }
            },
            'executorLabel': 'exec-importer'
        }
        component_spec = pipeline_spec_pb2.ComponentSpec()
        json_format.ParseDict(component_spec_dict, component_spec)

        executor_spec_dict = {
            'importer': {
                'artifactUri': {
                    'runtimeParameter': 'uri'
                },
                'typeSchema': {
                    'schemaTitle': 'system.Dataset',
                    'schemaVersion': '0.0.1'
                },
                'metadata': {
                    'foo': "{{$.inputs.parameters['metadata']}}"
                }
            }
        }
        executor_spec = pipeline_spec_pb2.PipelineDeploymentConfig.ExecutorSpec(
        )
        json_format.ParseDict(executor_spec_dict, executor_spec)

        outputs, task_status = importer_handler.run_importer(
            pipeline_resource_name='my-pipeline-2024-01-24-15-16-30-586674',
            component_name='comp-importer',
            component_spec=component_spec,
            executor_spec=executor_spec,
            arguments={
                'metadata': 'bar',
                'uri': '/fizz/buzz'
            },
            pipeline_root='/foo/bar',
            unique_pipeline_id='19024073',
        )
        expected_artifact = dsl.Dataset(
            name='artifact',
            uri='/fizz/buzz',
            metadata={'foo': 'bar'},
        )
        self.assertEqual(outputs['artifact'].schema_title,
                         expected_artifact.schema_title)
        self.assertEqual(outputs['artifact'].name, expected_artifact.name)
        self.assertEqual(outputs['artifact'].uri, expected_artifact.uri)
        self.assertEqual(outputs['artifact'].metadata,
                         expected_artifact.metadata)
        self.assertEqual(task_status, status.Status.SUCCESS)

    def test_uri_constant(self):
        component_spec_dict = {
            'inputDefinitions': {
                'parameters': {
                    'metadata': {
                        'parameterType': 'STRING'
                    },
                    'uri': {
                        'parameterType': 'STRING'
                    }
                }
            },
            'outputDefinitions': {
                'artifacts': {
                    'artifact': {
                        'artifactType': {
                            'schemaTitle': 'system.Artifact',
                            'schemaVersion': '0.0.1'
                        }
                    }
                }
            },
            'executorLabel': 'exec-importer'
        }
        component_spec = pipeline_spec_pb2.ComponentSpec()
        json_format.ParseDict(component_spec_dict, component_spec)

        executor_spec_dict = {
            'importer': {
                'artifactUri': {
                    'constant': 'gs://path'
                },
                'typeSchema': {
                    'schemaTitle': 'system.Artifact',
                    'schemaVersion': '0.0.1'
                },
                'metadata': {
                    'foo': [
                        "{{$.inputs.parameters['metadata']}}",
                        "{{$.inputs.parameters['metadata']}}"
                    ]
                }
            }
        }
        executor_spec = pipeline_spec_pb2.PipelineDeploymentConfig.ExecutorSpec(
        )
        json_format.ParseDict(executor_spec_dict, executor_spec)

        outputs, task_status = importer_handler.run_importer(
            pipeline_resource_name='my-pipeline-2024-01-24-15-16-30-586674',
            component_name='comp-importer',
            component_spec=component_spec,
            executor_spec=executor_spec,
            arguments={
                'metadata': 'text',
                'uri': 'gs://path'
            },
            pipeline_root='/foo/bar',
            unique_pipeline_id='19024073',
        )
        expected_artifact = dsl.Artifact(
            name='artifact',
            uri='gs://path',
            metadata={'foo': ['text', 'text']},
        )
        self.assertEqual(outputs['artifact'].schema_title,
                         expected_artifact.schema_title)
        self.assertEqual(outputs['artifact'].name, expected_artifact.name)
        self.assertEqual(outputs['artifact'].uri, expected_artifact.uri)
        self.assertEqual(outputs['artifact'].metadata,
                         expected_artifact.metadata)
        self.assertEqual(task_status, status.Status.SUCCESS)


class TestGetImporterUri(unittest.TestCase):

    def test_constant(self):
        importer_spec_dict = {
            'artifactUri': {
                'constant': '/foo/path'
            },
            'typeSchema': {
                'schemaTitle': 'system.Artifact',
                'schemaVersion': '0.0.1'
            },
            'metadata': {
                'foo': 'bar'
            }
        }
        importer_spec = pipeline_spec_pb2.PipelineDeploymentConfig.ImporterSpec(
        )
        json_format.ParseDict(importer_spec_dict, importer_spec)
        executor_input_dict = {
            'inputs': {
                'parameterValues': {
                    'uri': '/foo/path'
                }
            },
            'outputs': {
                'artifacts': {
                    'artifact': {
                        'artifacts': [{
                            'name': 'artifact',
                            'type': {
                                'schemaTitle': 'system.Artifact',
                                'schemaVersion': '0.0.1'
                            },
                            'uri': '/pipeline_root/task/artifact',
                            'metadata': {}
                        }]
                    }
                },
                'outputFile': '/pipeline_root/task/executor_output.json'
            }
        }
        executor_input = pipeline_spec_pb2.ExecutorInput()
        json_format.ParseDict(executor_input_dict, executor_input)
        uri = importer_handler.get_importer_uri(
            importer_spec=importer_spec,
            executor_input=executor_input,
        )
        self.assertEqual(uri, '/foo/path')

    def test_runtime_parameter(self):
        importer_spec_dict = {
            'artifactUri': {
                'runtimeParameter': 'uri'
            },
            'typeSchema': {
                'schemaTitle': 'system.Artifact',
                'schemaVersion': '0.0.1'
            },
            'metadata': {
                'foo': 'bar'
            }
        }
        importer_spec = pipeline_spec_pb2.PipelineDeploymentConfig.ImporterSpec(
        )
        json_format.ParseDict(importer_spec_dict, importer_spec)
        executor_input_dict = {
            'inputs': {
                'parameterValues': {
                    'uri': '/fizz/buzz'
                }
            },
            'outputs': {
                'artifacts': {
                    'artifact': {
                        'artifacts': [{
                            'name': 'artifact',
                            'type': {
                                'schemaTitle': 'system.Artifact',
                                'schemaVersion': '0.0.1'
                            },
                            'uri': '/pipeline_root/foo/task/artifact',
                            'metadata': {}
                        }]
                    }
                },
                'outputFile': '/pipeline_root/task/executor_output.json'
            }
        }
        executor_input = pipeline_spec_pb2.ExecutorInput()
        json_format.ParseDict(executor_input_dict, executor_input)
        uri = importer_handler.get_importer_uri(
            importer_spec=importer_spec,
            executor_input=executor_input,
        )
        self.assertEqual(uri, '/fizz/buzz')

    def test_constant_warns(self):
        importer_spec_dict = {
            'artifactUri': {
                'constant': 'gs://foo/bar'
            },
            'typeSchema': {
                'schemaTitle': 'system.Artifact',
                'schemaVersion': '0.0.1'
            },
            'metadata': {
                'foo': 'bar'
            }
        }
        importer_spec = pipeline_spec_pb2.PipelineDeploymentConfig.ImporterSpec(
        )
        json_format.ParseDict(importer_spec_dict, importer_spec)
        executor_input_dict = {
            'inputs': {
                'parameterValues': {
                    'uri': 'gs://foo/bar'
                }
            },
            'outputs': {
                'artifacts': {
                    'artifact': {
                        'artifacts': [{
                            'name': 'artifact',
                            'type': {
                                'schemaTitle': 'system.Artifact',
                                'schemaVersion': '0.0.1'
                            },
                            'uri': '/pipeline_root/foo/task/artifact',
                            'metadata': {}
                        }]
                    }
                },
                'outputFile': '/pipeline_root/task/executor_output.json'
            }
        }
        executor_input = pipeline_spec_pb2.ExecutorInput()
        json_format.ParseDict(executor_input_dict, executor_input)
        with self.assertWarnsRegex(
                UserWarning,
                r"It looks like you're using the remote file 'gs://foo/bar' in a 'dsl\.importer'\. Note that you will only be able to read and write to/from local files using 'artifact\.path' in local executed pipelines\."
        ):
            uri = importer_handler.get_importer_uri(
                importer_spec=importer_spec,
                executor_input=executor_input,
            )
        self.assertEqual(uri, 'gs://foo/bar')

    def test_runtime_parameter_warns(self):
        importer_spec_dict = {
            'artifactUri': {
                'runtimeParameter': 'uri'
            },
            'typeSchema': {
                'schemaTitle': 'system.Artifact',
                'schemaVersion': '0.0.1'
            },
            'metadata': {
                'foo': 'bar'
            }
        }
        importer_spec = pipeline_spec_pb2.PipelineDeploymentConfig.ImporterSpec(
        )
        json_format.ParseDict(importer_spec_dict, importer_spec)
        executor_input_dict = {
            'inputs': {
                'parameterValues': {
                    'uri': 's3://fizz/buzz'
                }
            },
            'outputs': {
                'artifacts': {
                    'artifact': {
                        'artifacts': [{
                            'name': 'artifact',
                            'type': {
                                'schemaTitle': 'system.Artifact',
                                'schemaVersion': '0.0.1'
                            },
                            'uri': '/pipeline_root/foo/task/artifact',
                            'metadata': {}
                        }]
                    }
                },
                'outputFile': '/pipeline_root/task/executor_output.json'
            }
        }
        executor_input = pipeline_spec_pb2.ExecutorInput()
        json_format.ParseDict(executor_input_dict, executor_input)
        with self.assertWarnsRegex(
                UserWarning,
                r"It looks like you're using the remote file 's3://fizz/buzz' in a 'dsl\.importer'\. Note that you will only be able to read and write to/from local files using 'artifact\.path' in local executed pipelines\."
        ):
            uri = importer_handler.get_importer_uri(
                importer_spec=importer_spec,
                executor_input=executor_input,
            )
        self.assertEqual(uri, 's3://fizz/buzz')


class TestGetArtifactClassForSchemaTitle(unittest.TestCase):

    def test_artifact(self):
        actual = importer_handler.get_artifact_class_from_schema_title(
            'system.Artifact')
        expected = dsl.Artifact
        self.assertEqual(actual, expected)

    def test_classification_metrics(self):
        actual = importer_handler.get_artifact_class_from_schema_title(
            'system.ClassificationMetrics')
        expected = dsl.ClassificationMetrics
        self.assertEqual(actual, expected)

    def test_not_system_type(self):
        actual = importer_handler.get_artifact_class_from_schema_title(
            'unknown.Type')
        expected = dsl.Artifact
        self.assertEqual(actual, expected)


if __name__ == '__main__':
    unittest.main()



# Copyright 2023 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for constructing the ExecutorInput message."""
import datetime
import os
from typing import Any, Dict

from google.protobuf import json_format
from google.protobuf import struct_pb2
from kfp import dsl
from kfp.compiler import pipeline_spec_builder
from kfp.dsl import utils
from kfp.pipeline_spec import pipeline_spec_pb2

_EXECUTOR_OUTPUT_FILE = 'executor_output.json'


def construct_executor_input(
    component_spec: pipeline_spec_pb2.ComponentSpec,
    arguments: Dict[str, Any],
    task_root: str,
    block_input_artifact: bool,
) -> pipeline_spec_pb2.ExecutorInput:
    """Constructs the executor input message for a task execution."""
    input_parameter_keys = list(
        component_spec.input_definitions.parameters.keys())
    # need to also add injected input parameters for f-string
    input_parameter_keys += [
        k for k, v in arguments.items() if not isinstance(v, dsl.Artifact)
    ]
    input_artifact_keys = list(
        component_spec.input_definitions.artifacts.keys())
    if input_artifact_keys and block_input_artifact:
        raise ValueError(
            'Input artifacts are not yet supported for local execution.')

    inputs = pipeline_spec_pb2.ExecutorInput.Inputs(
        parameter_values={
            param_name:
            pipeline_spec_builder.to_protobuf_value(arguments[param_name])
            if param_name in arguments else component_spec.input_definitions
            .parameters[param_name].default_value
            for param_name in input_parameter_keys
        },
        # input artifact constants are not supported yet,
        # except when passed from an upstream output or parent component input
        artifacts={
            artifact_name:
            dsl_artifact_to_artifact_list(arguments[artifact_name])
            for artifact_name, _ in
            component_spec.input_definitions.artifacts.items()
        },
    )

    output_parameter_keys = list(
        component_spec.output_definitions.parameters.keys())
    outputs = pipeline_spec_pb2.ExecutorInput.Outputs(
        parameters={
            param_name: pipeline_spec_pb2.ExecutorInput.OutputParameter(
                output_file=os.path.join(task_root, param_name))
            for param_name in output_parameter_keys
        },
        artifacts={
            artifact_name: artifact_type_schema_to_artifact_list(
                name=artifact_name,
                artifact_type=artifact_spec.artifact_type,
                task_root=task_root,
            ) for artifact_name, artifact_spec in
            component_spec.output_definitions.artifacts.items()
        },
        output_file=os.path.join(task_root, _EXECUTOR_OUTPUT_FILE),
    )
    return pipeline_spec_pb2.ExecutorInput(
        inputs=inputs,
        outputs=outputs,
    )


def get_local_pipeline_resource_name(pipeline_name: str) -> str:
    """Gets the local pipeline resource name from the pipeline name in
    PipelineSpec.

    Args:
        pipeline_name: The pipeline name provided by PipelineSpec.pipelineInfo.name.

    Returns:
        The local pipeline resource name. Includes timestamp.
    """
    timestamp = datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S-%f')
    return f'{pipeline_name}-{timestamp}'


def get_local_task_resource_name(component_name: str) -> str:
    """Gets the local task resource name from the component name in
    PipelineSpec.

    Args:
        component_name: The component name provided as the key for the component's ComponentSpec
    message. Takes the form comp-*.

    Returns:
        The local task resource name.
    """
    return component_name[len(utils.COMPONENT_NAME_PREFIX):]


def construct_local_task_root(
    pipeline_root: str,
    pipeline_resource_name: str,
    task_resource_name: str,
) -> str:
    """Constructs the local task root directory for a task."""
    return os.path.join(
        pipeline_root,
        pipeline_resource_name,
        task_resource_name,
    )


def artifact_type_schema_to_artifact_list(
    name: str,
    artifact_type: pipeline_spec_pb2.ArtifactTypeSchema,
    task_root: str,
) -> pipeline_spec_pb2.ArtifactList:
    """Constructs an ArtifactList instance for an artifact in ExecutorInput."""
    return pipeline_spec_pb2.ArtifactList(artifacts=[
        pipeline_spec_pb2.RuntimeArtifact(
            name=name,
            type=artifact_type,
            uri=os.path.join(task_root, name),
            # metadata always starts empty for output artifacts
            metadata=struct_pb2.Struct(),
        )
    ])


def dict_to_protobuf_struct(d: Dict[str, Any]) -> struct_pb2.Struct:
    """Converts a Python dictionary to a prototobuf Struct."""
    protobuf_struct = struct_pb2.Struct()
    protobuf_struct.update(d)
    return protobuf_struct


def dsl_artifact_to_artifact_list(
        artifact: dsl.Artifact) -> pipeline_spec_pb2.ArtifactList:
    """Converts a single dsl.Aritfact to a protobuf ArtifactList."""
    return pipeline_spec_pb2.ArtifactList(artifacts=[
        pipeline_spec_pb2.RuntimeArtifact(
            name=artifact.name,
            type=pipeline_spec_pb2.ArtifactTypeSchema(
                schema_title=artifact.schema_title,
                schema_version=artifact.schema_version),
            uri=artifact.uri,
            metadata=dict_to_protobuf_struct(artifact.metadata),
        )
    ])


def executor_input_to_dict(
    executor_input: pipeline_spec_pb2.ExecutorInput,
    component_spec: pipeline_spec_pb2.ComponentSpec,
) -> Dict[str, Any]:
    """Converts the executor input to a dictionary.

    Since protobuf value represents ints and floats the same way, we
    cast ints to their correct type. This should be called before
    replacing placeholders with values.

    This is consistent with the remote backend behavior.
    """
    executor_input_dict = json_format.MessageToDict(executor_input)
    inputs_typed_int = [
        in_param_name for in_param_name, parameter_spec in
        component_spec.input_definitions.parameters.items()
        if parameter_spec.parameter_type ==
        pipeline_spec_pb2.ParameterType.ParameterTypeEnum.NUMBER_INTEGER
    ]
    for param_name, param_value in executor_input_dict.get('inputs', {}).get(
            'parameterValues', {}).items():
        if param_name in inputs_typed_int:
            executor_input_dict['inputs']['parameterValues'][param_name] = int(
                param_value)
    return executor_input_dict



# Copyright 2023 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Objects for configuring local execution."""
import abc
import dataclasses
import os
from typing import Union

from kfp import local


class LocalRunnerType(abc.ABC):
    """The ABC for user-facing Runner configurations.

    Subclasses should be a dataclass.

    They should implement a .validate() method.
    """

    @abc.abstractmethod
    def validate(self) -> None:
        """Validates that the configuration arguments provided by the user are
        valid."""
        raise NotImplementedError


@dataclasses.dataclass
class SubprocessRunner:
    """Runner that indicates that local tasks should be run in a subprocess.

    Args:
        use_venv: Whether to run the subprocess in a virtual environment. If True, dependencies will be installed in the virtual environment. If False, dependencies will be installed in the current environment. Using a virtual environment is recommended.
    """
    use_venv: bool = True


class DockerRunner:
    """Runner that indicates that local tasks should be run as a Docker
    container."""
    DOCKER_CONTAINER_RUN_ARGS = {
        # for available run parameters:
        # https://docker-py.readthedocs.io/en/stable/containers.html
        'image',
        'command',
        'auto_remove',
        'blkio_weight_device',
        'blkio_weight',
        'cap_add',
        'cap_drop',
        'cgroup_parent',
        'cgroupns',
        'cpu_count',
        'cpu_percent',
        'cpu_period',
        'cpu_quota',
        'cpu_rt_period',
        'cpu_rt_runtime',
        'cpu_shares',
        'cpuset_cpus',
        'cpuset_mems',
        'detach',
        'device_cgroup_rules',
        'device_read_bps',
        'device_read_iops',
        'device_write_bps',
        'device_write_iops',
        'devices',
        'device_requests',
        'dns',
        'dns_opt',
        'dns_search',
        'domainname',
        'entrypoint',
        'environment',
        'extra_hosts',
        'group_add',
        'healthcheck',
        'hostname',
        'init',
        'init_path',
        'ipc_mode',
        'isolation',
        'kernel_memory',
        'labels',
        'links',
        'log_config',
        'lxc_conf',
        'mac_address',
        'mem_limit',
        'mem_reservation',
        'mem_swappiness',
        'memswap_limit',
        'mounts',
        'name',
        'nano_cpus',
        'network',
        'network_disabled',
        'network_mode',
        'networking_config',
        'oom_kill_disable',
        'oom_score_adj',
        'pid_mode',
        'pids_limit',
        'platform',
        'ports',
        'privileged',
        'publish_all_ports',
        'read_only',
        'remove',
        'restart_policy',
        'runtime',
        'security_opt',
        'shm_size',
        'stdin_open',
        'stdout',
        'stderr',
        'stop_signal',
        'storage_opt',
        'stream',
        'sysctls',
        'tmpfs',
        'tty',
        'ulimits',
        'use_config_proxy',
        'user',
        'userns_mode',
        'uts_mode',
        'version',
        'volume_driver',
        'volumes',
        'working_dir',
    }

    def __init__(self, **container_run_args):
        """Runner constructor, taking any arguments to propagate to
        `containers.run` in the `docker` SDK.

        Args:
            **container_run_args: Keyword arguments that comport with `containers.run` in the `docker` SDK, with some exceptions (see below).

        `containers.run` arguments are supported with the following exceptions:
        - image
        - command
        - volumes

        Raises:
            ImportError: Raised when `docker` is not installed
            ValueError: Raised when unsupported `containers.run` arguments are supplied
        """
        try:
            import docker  # noqa
        except ImportError as e:
            raise ImportError(
                f"Package 'docker' must be installed to use {DockerRunner.__name__!r}. Install it using 'pip install docker'."
            ) from e

        unsupported_args = container_run_args.keys(
        ) - self.DOCKER_CONTAINER_RUN_ARGS

        if unsupported_args:
            raise ValueError(
                f"Unsupported `docker run` arguments, see Package `docker` for details: {', '.join(unsupported_args)}"
            )

        excess_args: set = set(['image', 'command', 'volumes'
                               ]) & container_run_args.keys()

        if excess_args:
            raise ValueError(
                f"The following docker run arguments should not be specififed: {', '.join(excess_args)}"
            )

        self.container_run_args = container_run_args


class LocalExecutionConfig:
    instance = None

    def __new__(
        cls,
        runner: SubprocessRunner,
        pipeline_root: str,
        raise_on_error: bool,
    ) -> 'LocalExecutionConfig':
        # singleton pattern
        cls.instance = super(LocalExecutionConfig, cls).__new__(cls)
        return cls.instance

    def __init__(
        self,
        runner: SubprocessRunner,
        pipeline_root: str,
        raise_on_error: bool,
    ) -> None:
        permitted_runners = (SubprocessRunner, DockerRunner)
        if not isinstance(runner, permitted_runners):
            raise ValueError(
                f'Got unknown runner {runner} of type {runner.__class__.__name__}. Runner should be one of the following types: {". ".join(prunner.__name__ for prunner in permitted_runners)}.'
            )
        self.runner = runner
        self.pipeline_root = pipeline_root
        self.raise_on_error = raise_on_error

    @classmethod
    def validate(cls):
        if cls.instance is None:
            raise RuntimeError(
                f"Local environment not initialized. Please run '{local.__name__}.{init.__name__}()' before executing tasks locally."
            )


def init(
    # annotate with subclasses, not parent class, for more helpful ref docs
    runner: Union[SubprocessRunner, DockerRunner],
    pipeline_root: str = './local_outputs',
    raise_on_error: bool = True,
) -> None:
    """Initializes a local execution session.

    Once called, components can be invoked locally outside of a pipeline definition.

    Args:
        runner: The runner to use. Supported runners: kfp.local.SubprocessRunner and kfp.local.DockerRunner.
        pipeline_root: Destination for task outputs.
        raise_on_error: If True, raises an exception when a local task execution fails. If False, fails gracefully and does not terminate the current program.
    """
    # updates a global config
    pipeline_root = os.path.abspath(pipeline_root)
    LocalExecutionConfig(
        runner=runner,
        pipeline_root=pipeline_root,
        raise_on_error=raise_on_error,
    )
